from collections.abc import Iterable
import fnmatch

import openmdao.api as om
from openmdao.utils.array_utils import shape_to_len
from dymos.utils.misc import _unspecified
from ..phase.options import StateOptionsDictionary, TimeseriesOutputOptionsDictionary
from .misc import get_rate_units


def classify_var(var, state_options, parameter_options, control_options, polynomial_control_options):
    """
    Classifies a variable of the given name or path.

    This method searches for it as a time variable, state variable,
    control variable, or parameter.  If it is not found to be one
    of those variables, it is assumed to be the path to a variable
    relative to the top of the ODE system for the phase.

    Parameters
    ----------
    var : str
        The name of the variable to be classified.
    state_options : dict of {str: OptionsDictionary}
        For each state variable, a dictionary of its options, keyed by name.
    parameter_options : dict of {str: OptionsDictionary}
        For each parameter, a dictionary of its options, keyed by name.
    control_options : dict of {str: OptionsDictionary}
        For each control variable, a dictionary of its options, keyed by name.
    polynomial_control_options : dict of {str: OptionsDictionary}
        For each polynomial variable, a dictionary of its options, keyed by name.

    Returns
    -------
    str
        The classification of the given variable, which is one of
        'time', 'time_phase', 'state', 'input_control', 'indep_control', 'control_rate',
        'control_rate2', 'input_polynomial_control', 'indep_polynomial_control',
        'polynomial_control_rate', 'polynomial_control_rate2', 'parameter',
        or 'ode'.
    """
    if var in {'time', 'time_phase'}:
        return var
    elif var in state_options:
        return 'state'
    elif var in control_options:
        if control_options[var]['opt']:
            return 'indep_control'
        else:
            return 'input_control'
    elif var in polynomial_control_options:
        if polynomial_control_options[var]['opt']:
            return 'indep_polynomial_control'
        else:
            return 'input_polynomial_control'
    elif var in parameter_options:
        return 'parameter'
    elif var.endswith('_rate'):
        if var[:-5] in control_options:
            return 'control_rate'
        elif var[:-5] in polynomial_control_options:
            return 'polynomial_control_rate'
    elif var.endswith('_rate2'):
        if var[:-6] in control_options:
            return 'control_rate2'
        elif var[:-6] in polynomial_control_options:
            return 'polynomial_control_rate2'

    return 'ode'


def get_promoted_vars(ode, iotypes, metadata_keys=None, get_remote=True):
    """
    Returns a dictionary mapping the promoted names of all inputs in a system to their associated metadata.

    Parameters
    ----------
    ode : openmdao.core.System
        The system from which the promoted inputs or outputs are being retrieved.
    iotypes : str or tuple
        One of 'input' or 'output', or a tuple of both.
    metadata_keys : Iterable or None
        Additional metadata requested for the variables. By default returns metadata available on 'allprocs'.  See
        openmdao.core.System.get_io_metadata for more information.
    get_remote : bool
        If True, include IO not local to this proc.

    Returns
    -------
    dict
        A dictionary mapping the promoted names of inputs in the system to their associated metadata.
    """
    _iotypes = (iotypes,) if isinstance(iotypes, str) else iotypes
    return {opts['prom_name']: opts for opts in ode.get_io_metadata(iotypes=_iotypes, get_remote=get_remote,
                                                                    metadata_keys=metadata_keys).values()}


def get_targets(ode, name, user_targets, control_rates=False):
    """
    Return the targets of a variable in a given ODE system.

    If the targets of the variable is _unspecified, and the name is a top level input name
    in the ODE, then the values are automatically connected to that top-level input.
    If _unspecified and not a top-level input of the ODE, no connection is made.
    If targets is explicitly None, then no connection is made.
    Otherwise, if the user specified some other string or sequence of strings as targets, then
    those are returned.

    Parameters
    ----------
    ode : openmdao.core.System or dict
        The OpenMDAO system which serves as the ODE for dymos, or a dictionary of promoted ODE input names and
        their associated metadata. If a System, it should already have had its setup and configure methods called.
    name : str
        The name of the variable whose targets are desired.
    user_targets : str or None or Sequence or _unspecified
        Targets for the variable as given by the user.
    control_rates : bool or int
        If True, search for the target of the variable with the given name.  If 1, search for
        the first rate of the variable '{control_name}_rate', and if 2, search for the second
        derivative of the variable '{control_name}_rate2'.

    Returns
    -------
    list
        The target inputs of the variable in the ODE, as a list.

    Notes
    -----
    This method requires that the ODE has run its setup and configure methods.  Thus,
    this method should be called from configure of some parent Group, and the ODE should
    be a system within that Group.
    """
    if isinstance(ode, dict):
        ode_inputs = ode
    else:
        ode_inputs = {opts['prom_name']: opts for opts in
                      ode.get_io_metadata(iotypes=('input',), get_remote=True).values()}
    if user_targets is _unspecified:
        if name in ode_inputs and control_rates not in {1, 2}:
            return [name]
        elif control_rates == 1 and f'{name}_rate' in ode_inputs:
            return [f'{name}_rate']
        elif control_rates == 2 and f'{name}_rate2' in ode_inputs:
            return [f'{name}_rate2']
    elif user_targets:
        if isinstance(user_targets, str):
            return [user_targets]
        else:
            return user_targets

    return []


def _configure_constraint_introspection(phase):
    """
    Modify constraint options in-place using introspection of the phase and its ODE.

    Parameters
    ----------
    phase : Phase
        The phase object whose boundary and path constraints are to be introspected.
    """
    for constraint_type, constraints in [('initial', phase._initial_boundary_constraints),
                                         ('final', phase._final_boundary_constraints),
                                         ('path', phase._path_constraints)]:
        for con in constraints:
            time_units = phase.time_options['units']

            # Determine the path to the variable which we will be constraining
            var = con['name']
            var_type = phase.classify_var(var)

            if var != con['constraint_name'] is not None and var_type != 'ode':
                om.issue_warning(f"Option 'constraint_name' on {constraint_type} constraint {var} is only "
                                 f"valid for ODE outputs. The option is being ignored.", om.UnusedOptionWarning)

            if var_type in {'time', 'time_phase'}:
                con['shape'] = (1,)
                con['units'] = time_units if con['units'] is None else con['units']
                con['constraint_path'] = f'timeseries.{var_type}'

            elif var_type == 'state':
                state_shape = phase.state_options[var]['shape']
                state_units = phase.state_options[var]['units']
                con['shape'] = state_shape
                con['units'] = state_units if con['units'] is None else con['units']
                con['constraint_path'] = f'timeseries.states:{var}'

            elif var_type == 'parameter':
                param_shape = phase.parameter_options[var]['shape']
                param_units = phase.parameter_options[var]['units']
                con['shape'] = param_shape
                con['units'] = param_units if con['units'] is None else con['units']
                con['constraint_path'] = f'parameter_vals:{var}'

            elif var_type == 'indep_control':
                control_shape = phase.control_options[var]['shape']
                control_units = phase.control_options[var]['units']

                con['shape'] = control_shape
                con['units'] = control_units if con['units'] is None else con['units']
                con['constraint_path'] = f'timeseries.controls:{var}'

            elif var_type == 'input_control':
                control_shape = phase.control_options[var]['shape']
                control_units = phase.control_options[var]['units']

                con['shape'] = control_shape
                con['units'] = control_units if con['units'] is None else con['units']
                con['constraint_path'] = f'timeseries.controls:{var}'

            elif var_type == 'indep_polynomial_control':
                control_shape = phase.polynomial_control_options[var]['shape']
                control_units = phase.polynomial_control_options[var]['units']
                con['shape'] = control_shape
                con['units'] = control_units if con['units'] is None else con['units']
                con['constraint_path'] = f'timeseries.polynomial_controls:{var}'

            elif var_type == 'input_polynomial_control':
                control_shape = phase.polynomial_control_options[var]['shape']
                control_units = phase.polynomial_control_options[var]['units']
                con['shape'] = control_shape
                con['units'] = control_units if con['units'] is None else con['units']
                con['constraint_path'] = f'timeseries.polynomial_controls:{var}'

            elif var_type == 'control_rate':
                control_name = var[:-5]
                control_shape = phase.control_options[control_name]['shape']
                control_units = phase.control_options[control_name]['units']
                con['shape'] = control_shape
                con['units'] = get_rate_units(control_units, time_units, deriv=1) \
                    if con['units'] is None else con['units']
                con['constraint_path'] = f'timeseries.control_rates:{var}'

            elif var_type == 'control_rate2':
                control_name = var[:-6]
                control_shape = phase.control_options[control_name]['shape']
                control_units = phase.control_options[control_name]['units']
                con['shape'] = control_shape
                con['units'] = get_rate_units(control_units, time_units, deriv=2) \
                    if con['units'] is None else con['units']
                con['constraint_path'] = f'timeseries.control_rates:{var}'

            elif var_type == 'polynomial_control_rate':
                control_name = var[:-5]
                control_shape = phase.polynomial_control_options[control_name]['shape']
                control_units = phase.polynomial_control_options[control_name]['units']
                con['shape'] = control_shape
                con['units'] = get_rate_units(control_units, time_units, deriv=1) \
                    if con['units'] is None else con['units']
                con['constraint_path'] = f'timeseries.polynomial_control_rates:{var}'

            elif var_type == 'polynomial_control_rate2':
                control_name = var[:-6]
                control_shape = phase.polynomial_control_options[control_name]['shape']
                control_units = phase.polynomial_control_options[control_name]['units']
                con['shape'] = control_shape
                con['units'] = get_rate_units(control_units, time_units, deriv=2) \
                    if con['units'] is None else con['units']
                con['constraint_path'] = f'timeseries.polynomial_control_rates:{var}'

            else:
                # Failed to find variable, assume it is in the ODE. This requires introspection.
                ode = phase.options['transcription']._get_ode(phase)

                meta = get_source_metadata(ode, src=var, user_units=con['units'], user_shape=con['shape'])

                con['shape'] = meta['shape']
                con['units'] = meta['units']
                con['constraint_path'] = f'timeseries.{con["constraint_name"]}'


def configure_controls_introspection(control_options, ode, time_units='s'):
    """
    Modify control options in-place using introspection of the user-provided ODE.

    Parameters
    ----------
    control_options : dict of {str: ControlOptionsDictionary
        A dictionary keyed by control name containing the options for all controls to be applied
        to the ODE.
    ode : om.System
        An instantiated System that serves as the ODE to which the controls should be applied.
    time_units : str
        The units of time for the Phase.

    Raises
    ------
    ValueError
        If the control or one of its rates are connected to a variable that is tagged as static
        within the ODE.
    """
    ode_inputs = get_promoted_vars(ode, iotypes='input')
    for name, options in control_options.items():

        targets, shape, units, static_target = _get_targets_metadata(ode_inputs,
                                                                     name=name,
                                                                     user_targets=options['targets'],
                                                                     user_units=options['units'],
                                                                     user_shape=options['shape'],
                                                                     control_rate=False)
        options['targets'] = targets
        if targets:
            options['units'] = units
            options['shape'] = shape

        if static_target:
            raise ValueError(f"Control '{name}' cannot be connected to its targets because one "
                             f"or more targets are tagged with 'dymos.static_target'.")

        # Now check rate targets
        rate_targets, shape, units, static_target = _get_targets_metadata(ode_inputs,
                                                                          name=name,
                                                                          user_targets=options['rate_targets'],
                                                                          user_units=options['units'],
                                                                          user_shape=options['shape'],
                                                                          control_rate=1)
        options['rate_targets'] = rate_targets
        if options['units'] is _unspecified:
            options['units'] = time_units if units is None else f'{units}*{time_units}'
        options['shape'] = shape
        if static_target:
            raise ValueError(f"Control rate of '{name}' cannot be connected to its targets "
                             f"because one or more targets are tagged with 'dymos.static_target'.")

        # Now check rate2 targets
        rate2_targets, shape, units, static_target = _get_targets_metadata(ode_inputs,
                                                                           name=name,
                                                                           user_targets=options['rate2_targets'],
                                                                           user_units=options['units'],
                                                                           user_shape=options['shape'],
                                                                           control_rate=2)
        options['rate2_targets'] = rate2_targets
        if options['units'] is _unspecified:
            options['units'] = time_units if units is None else f'{units}*{time_units}**2'
        options['shape'] = shape
        if static_target:
            raise ValueError(f"Control rate2 of '{name}' cannot be connected to its targets "
                             f"because one or more targets are tagged with 'dymos.static_target'.")


def configure_parameters_introspection(parameter_options, ode):
    """
    Modify parameter options in-place using introspection of the user-provided ODE.

    Parameters
    ----------
    parameter_options : dict of {str: ParameterOptionsDictionary
        A dictionary keyed by parameter name containing the options for all parameters to be applied
        to the ODE. Options for 'targets', 'units', 'shape', and 'static_target' are modified in-place.
    ode : om.System
        An instantiated System that serves as the ODE to which the parameters should be applied.
    """
    ode_inputs = get_promoted_vars(ode, iotypes='input')

    for name, options in parameter_options.items():
        try:
            targets, shape, units, static_target = _get_targets_metadata(ode_inputs, name=name,
                                                                         user_targets=options['targets'],
                                                                         user_units=options['units'],
                                                                         user_shape=options['shape'],
                                                                         user_static_target=options['static_target'])
        except ValueError as e:
            raise ValueError(f'Parameter `{name}` has invalid target(s).\n{str(e)}') from e
        options['targets'] = targets
        options['units'] = units
        options['shape'] = shape
        options['static_target'] = static_target


def configure_time_introspection(time_options, ode):
    """
    Modify time options in-place using introspection of the user-provided ODE.

    Parameters
    ----------
    time_options : dict of {str: TimeOptionsDictionary
        A dictionary keyed by control name containing the options for all controls to be applied
        to the ODE. Options for 'targets', 'time_phase_targets', and 'units' are modified in-place.
    ode : om.System
        An instantiated System that serves as the ODE to which the controls should be applied.

    Raises
    ------
    ValueError
        If time or time_phase are connected to a variable that is tagged as static
        within the ODE.
    """
    ode_inputs = get_promoted_vars(ode, 'input')
    # time
    targets, shape, units, static_target = _get_targets_metadata(ode_inputs,
                                                                 name='time',
                                                                 user_targets=time_options['targets'],
                                                                 user_units=time_options['units'],
                                                                 user_shape=(1,))

    time_options['targets'] = targets
    time_options['units'] = units

    if static_target:
        raise ValueError(f"'time' cannot be connected to its targets because one "
                         f"or more targets are tagged with 'dymos.static_target'.")

    # time_phase
    targets, shape, units, static_target = _get_targets_metadata(ode_inputs,
                                                                 name='time_phase',
                                                                 user_targets=time_options['time_phase_targets'],
                                                                 user_units=time_options['units'],
                                                                 user_shape=(1,))

    time_options['time_phase_targets'] = targets

    if static_target:
        raise ValueError(f"'time_phase' cannot be connected to its targets because one "
                         f"or more targets are tagged with 'dymos.static_target'.")


def configure_states_introspection(state_options, time_options, control_options, parameter_options,
                                   polynomial_control_options, ode):
    """
    Modifies state options in-place, automatically determining 'targets', 'units', and 'shape' if necessary.

    The precedence rules for the state shape and units are as follows:
    1. If the user has specified units and shape in the state options, use those.
    2a. If the user has not specified shape, and targets exist, then pull the shape from the targets.
    2b. If the user has not specified shape and no targets exist, then pull the shape from the rate source.
    2c. If shape cannot be inferred, assume (1,)
    3a. If the user has not specified units, first try to pull units from a target
    3b. If there are no targets, pull units from the rate source and multiply by time units.

    Parameters
    ----------
    state_options : dict of {str: StateOptionsDictionary}
        The state variables to be configured.
    time_options : TimeOptionsDictionary
        The time options.
    control_options : dict of {str: ControlOptionsDictionary}
        The options for each control.
    parameter_options : dict of {str: ParameterOptionsDictionary}
        The options for each parameter.
    polynomial_control_options : dict of {str: PolynomialControlOptionsDictionary}
        The options for each polynomial control.
    ode : System
        The OpenMDAO system which provides the state rates as outputs.
    """
    time_units = time_options['units']
    ode_inputs = get_promoted_vars(ode, 'input')
    ode_outputs = get_promoted_vars(ode, 'output')

    for state_name, options in state_options.items():
        # Automatically determine targets of state if left _unspecified
        targets, tgt_shape, tgt_units, static_target = _get_targets_metadata(ode_inputs,
                                                                             state_name,
                                                                             user_targets=options['targets'],
                                                                             user_units=options['units'],
                                                                             user_shape=options['shape'])

        options['targets'] = targets
        if targets:
            options['shape'] = tgt_shape
            options['units'] = tgt_units

        # 3. Attempt rate-source introspection
        rate_src = options['rate_source']
        rate_src_type = classify_var(rate_src, state_options, parameter_options, control_options,
                                     polynomial_control_options)

        if rate_src_type in {'time', 'time_phase'}:
            rate_src_units = time_options['units']
            rate_src_shape = (1,)
        elif rate_src_type == 'state':
            rate_src_units = state_options[rate_src]['units']
            rate_src_shape = state_options[rate_src]['shape']
        elif rate_src_type in ['input_control', 'indep_control']:
            rate_src_units = control_options[rate_src]['units']
            rate_src_shape = control_options[rate_src]['shape']
        elif rate_src_type in ['input_polynomial_control', 'indep_polynomial_control']:
            rate_src_units = polynomial_control_options[rate_src]['units']
            rate_src_shape = polynomial_control_options[rate_src]['shape']
        elif rate_src_type == 'parameter':
            rate_src_units = parameter_options[rate_src]['units']
            rate_src_shape = parameter_options[rate_src]['shape']
        elif rate_src_type == 'control_rate':
            control_name = rate_src[:-5]
            control = control_options[control_name]
            rate_src_units = get_rate_units(control['units'], time_units, deriv=1)
            rate_src_shape = control['shape']
        elif rate_src_type == 'control_rate2':
            control_name = rate_src[:-6]
            control = control_options[control_name]
            rate_src_units = get_rate_units(control['units'], time_units, deriv=2)
            rate_src_shape = control['shape']
        elif rate_src_type == 'polynomial_control_rate':
            control_name = rate_src[:-5]
            control = polynomial_control_options[control_name]
            rate_src_units = get_rate_units(control['units'], time_units, deriv=1)
            rate_src_shape = control['shape']
        elif rate_src_type == 'polynomial_control_rate2':
            control_name = rate_src[:-6]
            control = polynomial_control_options[control_name]
            rate_src_units = get_rate_units(control['units'], time_units, deriv=2)
            rate_src_shape = control['shape']
        elif rate_src_type == 'ode':
            meta = get_source_metadata(ode_outputs, src=rate_src, user_units=options['units'], user_shape=options['shape'])
            rate_src_shape = meta['shape']
            rate_src_units = meta['units']
        else:
            rate_src_shape = (1,)
            rate_src_units = None

        if options['shape'] in {None, _unspecified}:
            options['shape'] = rate_src_shape

        if options['units'] is _unspecified:
            options['units'] = time_units if rate_src_units is None else f'{rate_src_units}*{time_units}'


def configure_analytic_states_introspection(state_options, ode):
    """
    Modifies state options in-place, automatically determining 'targets', 'units', and 'shape' if necessary.

    The precedence rules for the state shape and units are as follows:
    1. If the user has specified units and shape in the state options, use those.
    2a. If the user has not specified shape, and targets exist, then pull the shape from the targets.
    2b. If the user has not specified shape and no targets exist, then pull the shape from the rate source.
    2c. If shape cannot be inferred, assume (1,)
    3a. If the user has not specified units, first try to pull units from a target
    3b. If there are no targets, pull units from the rate source and multiply by time units.

    Parameters
    ----------
    state_options : dict of {str: StateOptionsDictionary}
        The state variables to be configured.
    ode : System
        The OpenMDAO system which provides the state rates as outputs.
    """
    ode_outputs = get_promoted_vars(ode, 'output')

    for state_name, options in state_options.items():
        # Automatically determine targets of state if left _unspecified
        source = options['source'] if options['source'] else state_name
        meta = get_source_metadata(ode_outputs, src=source, user_units=options['units'], user_shape=options['shape'])
        src_shape = meta['shape']
        src_units = meta['units']

        if 'dymos.static_output' in meta['tags']:
            raise RuntimeError(f'ODE output {source} is tagged with `dymos.static_output` and cannot be used as a '
                               f'state variable in an AnalyticPhase.')

        if options['shape'] in {None, _unspecified}:
            options['shape'] = src_shape

        if options['units'] is _unspecified:
            options['units'] = src_units


def configure_states_discovery(state_options, ode):
    """
    Searches phase output metadata for any declared states and adds them.

    Parameters
    ----------
    state_options : dict
        The dictionary of options for each state in the phase.
    ode : System
        The System instance providing the ODE for the phase.
    """
    out_meta = ode.get_io_metadata(iotypes='output', metadata_keys=['tags'],
                                   get_remote=True)

    for name, meta in out_meta.items():
        tags = meta['tags']
        prom_name = meta['prom_name']
        state = None
        for tag in sorted(tags):

            # Declared as rate_source.
            if tag.startswith('dymos.state_rate_source:') or tag.startswith('state_rate_source:'):
                state = tag.rpartition(':')[-1]
                if tag.startswith('state_rate_source:'):
                    msg = f"The tag '{tag}' has a deprecated format and will no longer work in " \
                          f"dymos version 2.0.0. Use 'dymos.state_rate_source:{state}' instead."
                    om.issue_warning(msg, category=om.OMDeprecationWarning)
                if state not in state_options:
                    state_options[state] = StateOptionsDictionary()
                    state_options[state]['name'] = state

                if state_options[state]['rate_source'] is not None:
                    if state_options[state]['rate_source'] != prom_name:
                        raise ValueError(f"rate_source has been declared twice for state "
                                         f"'{state}' which is tagged on '{name}'.")

                state_options[state]['rate_source'] = prom_name

            # Declares units for state.
            if tag.startswith('dymos.state_units:') or tag.startswith('state_units:'):
                tagged_state_units = tag.rpartition(':')[-1]
                if tag.startswith('state_units:'):
                    msg = f"The tag '{tag}' has a deprecated format and will no longer work in " \
                          f"dymos version 2.0.0. Use 'dymos.{tag}' instead."
                    om.issue_warning(msg, category=om.OMDeprecationWarning)
                if state is None:
                    raise ValueError(f"'{tag}' tag declared on '{prom_name}' also requires "
                                     f"that the 'dymos.state_rate_source:{tagged_state_units}' "
                                     f"tag be declared.")
                state_options[state]['units'] = tagged_state_units

    # Check over all existing states and make sure we aren't missing any rate sources.
    for name, options in state_options.items():
        if options['rate_source'] is None:
            raise ValueError(f"State '{name}' is missing a rate_source.")


def configure_analytic_states_discovery(state_options, ode):
    """
    Searches phase output metadata for any declared states and adds them.

    Parameters
    ----------
    state_options : dict
        The dictionary of options for each state in the phase.
    ode : System
        The System instance providing the ODE for the phase.
    """
    out_meta = ode.get_io_metadata(iotypes='output', metadata_keys=['tags'],
                                   get_remote=True)

    for name, meta in out_meta.items():
        tags = meta['tags']
        prom_name = meta['prom_name']
        state = None
        for tag in sorted(tags):

            # Declared as rate_source.
            if tag.startswith('dymos.state_source:'):
                state = tag.rpartition(':')[-1]
                if state not in state_options:
                    state_options[state] = StateOptionsDictionary()
                    state_options[state]['name'] = state

                if state_options[state]['source'] is not None:
                    if state_options[state]['source'] != prom_name:
                        raise ValueError(f"source has been declared twice for state "
                                         f"'{state}' which is tagged on '{name}'.")

                state_options[state]['source'] = prom_name

            # Declares units for state.
            if tag.startswith('dymos.state_units:') or tag.startswith('state_units:'):
                tagged_state_units = tag.rpartition(':')[-1]
                if state is None:
                    raise ValueError(f"'{tag}' tag declared on '{prom_name}' also requires "
                                     f"that the 'dymos.state_source:{tagged_state_units}' "
                                     f"tag be declared.")
                state_options[state]['units'] = tagged_state_units

    # Check over all existing states and make sure we aren't missing any rate sources.
    for name, options in state_options.items():
        if options['source'] is None:
            raise ValueError(f"State '{name}' is missing a source.")


def configure_timeseries_output_glob_expansion(phase):
    """
    Modify timeseries outputs in-place using introspection to expand any glob patterns.

    Parameters
    ----------
    phase : Phase
        The phase object whose boundary and path constraints are to be introspected.
    """
    transcription = phase.options['transcription']
    ode = transcription._get_ode(phase)
    ode_outputs = get_promoted_vars(ode, 'output')

    new_outputs = {}

    # Step 1: Expand globs
    for ts_name, ts_meta in phase._timeseries.items():

        explicit_requests = set([output['name'] for output in ts_meta['outputs'].values() if '*' not in output['name']])

        for output_name, output_options in ts_meta['outputs'].items():

            if '*' in output_name:
                matching_outputs = filter_outputs(output_name, ode_outputs)
                wildcard_units = {} if output_options['wildcard_units'] is None else output_options['wildcard_units']

                for op, meta in matching_outputs.items():
                    if op not in explicit_requests and 'dymos.static_output' not in meta['tags']:
                        new_output = TimeseriesOutputOptionsDictionary()
                        new_output['name'] = op
                        new_output['output_name'] = opname = op.split('.')[-1]
                        new_output['units'] = _unspecified if op not in wildcard_units else wildcard_units[op]
                        new_output['shape'] = _unspecified
                        if opname not in explicit_requests and opname in new_outputs:
                            raise RuntimeError(f'{phase.pathname}: The glob pattern `{output_name}` matches multiple '
                                               f'outputs in the ODE.\n'
                                               f'Add these outputs explicitly with unique output names using the\n'
                                               f'output_name argument to avoid this error.\n'
                                               f'Colliding names: {op}  {new_outputs[opname]["name"]}')
                        else:
                            new_outputs[new_output['output_name']] = new_output
            else:
                new_outputs[output_name] = output_options

        phase._timeseries[ts_name]['outputs'] = new_outputs


def configure_timeseries_output_introspection(phase):
    """
    Modify timeseries outputs in-place using introspection to find output units and shape.

    Parameters
    ----------
    phase : Phase
        The phase object whose timeseries outputs are to be introspected.
    """
    configure_timeseries_output_glob_expansion(phase)

    transcription = phase.options['transcription']

    for ts_name, ts_opts in phase._timeseries.items():

        not_found = set()

        for output_name, output_options in ts_opts['outputs'].items():
            try:
                output_meta = transcription._get_timeseries_var_source(output_options['name'],
                                                                       output_options['output_name'],
                                                                       phase=phase)
            except ValueError as e:
                not_found.add(output_name)

            output_options['src'] = output_meta['src']
            output_options['src_idxs'] = output_meta['src_idxs']

            if output_options['shape'] in (None, _unspecified):
                output_options['shape'] = output_meta['shape']

            if output_options['units'] in (None, _unspecified):
                output_options['units'] = output_meta['units']

        if not_found:
            sorted_list = ', '.join(sorted(not_found))
            om.issue_warning(f'{phase.pathname}: The following timeseries outputs were requested but not found in the '
                             f'ODE: {sorted_list}')

        for s in not_found:
            ts_opts['outputs'].pop(s)


def filter_outputs(patterns, sys):
    """
    Find all outputs of the given system that match one or more of the strings given in patterns.

    Parameters
    ----------
    patterns : str or Sequence
        A string or sequence of strings to be matched in the outputs of the given system.  These
        may include glob patterns.
    sys : System or dict
        The OpenMDAO system whose outputs are to be filtered, or a dictionary of outputs as returned by
        get_promoted_vars.

    Returns
    -------
    dict of {str: dict}
        A dictionary where the matching output names are the keys and the associated dict provides
        the 'units' and 'shapes' metadata.
    """
    outputs = sys if isinstance(sys, dict) else get_promoted_vars(sys, iotypes='output', metadata_keys=['shape', 'units', 'tags'])
    _patterns = [patterns] if isinstance(patterns, str) else patterns

    output_names = list(outputs.keys())
    filtered = set()
    results = {}

    for pattern in _patterns:
        filtered.update(fnmatch.filter(output_names, pattern))

    for var in filtered:
        results[var] = {'units': outputs[var]['units'], 'shape': outputs[var]['shape'], 'tags': outputs[var]['tags']}

    return results


def get_target_metadata(ode, name, user_targets=_unspecified, user_units=_unspecified,
                        user_shape=_unspecified, control_rate=False, user_static_target=_unspecified):
    """
    Return the targets of a state variable in a given ODE system.

    If the targets of the state is _unspecified, and the state name is a top level input name
    in the ODE, then the state values are automatically connected to that top-level input.
    If _unspecified and not a top-level input of the ODE, no connection is made.
    If targets is explicitly None, then no connection is made.
    Otherwise, if the user specified some other string or sequence of strings as targets, then
    those are returned.

    Parameters
    ----------
    ode : om.System or dict
        The OpenMDAO system which serves as the ODE for dymos, or a dictionary of inputs as returned by
        utils.introspection.get_promoted_vars.  If a system, it should already have had its setup and configure
        methods called.
    name : str
        The name of the variable whose targets are desired.
    user_targets : str or None or Sequence or _unspecified
        Targets for the variable as given by the user.
    user_units : str or None or _unspecified
        Units for the variable as given by the user.
    user_shape : None or Sequence or _unspecified
        Shape for the variable as given by the user.
    control_rate : bool
        When True, check for the control rate if the name is not in the ODE.
    user_static_target : bool or None or _unspecified
        When False, assume the shape of the target in the ODE includes the number of nodes as the
        first dimension.  If True, the connecting parameter does not need to be "fanned out" to
        connect to each node.  If _unspecified, attempt to resolve by the presence of a tag
        `dymos.static_target` on the target variable, which is the same as `static_target=True`.

    Returns
    -------
    shape : tuple
        The shape of the variable.  If not specified, shape is taken from the ODE targets.
    units : str
        The units of the variable.  If not specified, units are taken from the ODE targets.
    static_target : bool
        True if the target is static, otherwise False.

    Notes
    -----
    This method requires that the ODE has run its setup and configure methods.  Thus,
    this method should be called from configure of some parent Group, and the ODE should
    be a system within that Group.
    """
    ode_inputs = ode if isinstance(ode, dict) else get_promoted_vars(ode, iotypes='input')

    if user_targets is _unspecified:
        if name in ode_inputs:
            targets = [name]
        elif control_rate and f'{name}_rate' in ode_inputs:
            targets = [f'{name}_rate']
        elif control_rate and f'{name}_rate2' in ode_inputs:
            targets = [f'{name}_rate2']
        else:
            targets = []
    elif user_targets:
        if isinstance(user_targets, str):
            targets = [user_targets]
        else:
            targets = user_targets
    else:
        targets = []

    if user_units is _unspecified:
        target_units_set = {ode_inputs[tgt]['units'] for tgt in targets}
        if len(target_units_set) == 1:
            units = next(iter(target_units_set))
        else:
            raise ValueError(f'Unable to automatically assign units to {name}. '
                             f'Targets have multiple units: {target_units_set}. '
                             f'Either promote targets and use set_input_defaults to assign common '
                             f'units, or explicitly provide them to {name}.')
    else:
        units = user_units

    # Resolve whether the targets is static or dynamic
    static_target_tags = [tgt for tgt in targets if 'dymos.static_target' in ode_inputs[tgt]['tags']]
    if static_target_tags:
        static_target = True
        if not user_static_target:
            raise ValueError(f"User has specified 'static_target = False' for parameter {name},"
                             f"but one or more targets is tagged with "
                             f"'dymos.static_target': {' '.join(static_target_tags)}")
    else:
        if user_static_target is _unspecified:
            static_target = False
        else:
            static_target = user_static_target

    if user_shape in {None, _unspecified}:
        # Resolve target shape
        target_shape_set = {ode_inputs[tgt]['shape'] for tgt in targets}
        if len(target_shape_set) == 1:
            shape = next(iter(target_shape_set))
            if not static_target:
                if len(shape) == 1:
                    shape = (1,)
                else:
                    shape = shape[1:]
        elif len(target_shape_set) == 0:
            raise ValueError(f'Unable to automatically assign a shape to {name}.\n'
                             'Targets for this variable either do not exist or have no shape set.\n'
                             'The shape for this variable must be set explicitly via the '
                             '`shape=<tuple>` argument.')
        else:
            raise ValueError(f'Unable to automatically assign a shape to {name} based on targets. '
                             f'Targets have multiple shapes assigned: {target_shape_set}. '
                             f'Change targets such that all have common shapes.')
    else:
        shape = user_shape

    return shape, units, static_target


def _get_targets_metadata(ode, name, user_targets=_unspecified, user_units=_unspecified,
                          user_shape=_unspecified, control_rate=False, user_static_target=_unspecified):
    """
    Return the targets of a variable in a given ODE system and their metadata.

    If the targets of the state is _unspecified, and the state name is a top level input name
    in the ODE, then the state values are automatically connected to that top-level input.
    If _unspecified and not a top-level input of the ODE, no connection is made.
    If targets is explicitly None, then no connection is made.
    Otherwise, if the user specified some other string or sequence of strings as targets, then
    those are returned.

    Parameters
    ----------
    ode : om.System or dict
        The OpenMDAO system which serves as the ODE for dymos, or a dictionary of the ODE inputs and their metadata
        from a previous call to _get_targets_metadata. If given as the ODE system, it should already have had its setup
        and configure methods called.
    name : str
        The name of the variable whose targets are desired.
    user_targets : str or None or Sequence or _unspecified
        Targets for the variable as given by the user.
    user_units : str or None or _unspecified
        Units for the variable as given by the user.
    user_shape : None or Sequence or _unspecified
        Shape for the variable as given by the user.
    control_rate : bool
        When True, check for the control rate if the name is not in the ODE.
    user_static_target : bool or None or _unspecified
        When False, assume the shape of the target in the ODE includes the number of nodes as the
        first dimension.  If True, the connecting parameter does not need to be "fanned out" to
        connect to each node.  If _unspecified, attempt to resolve by the presence of a tag
        `dymos.static_target` on the target variable, which is the same as `static_target=True`.

    Returns
    -------
    targets : list
        The target inputs of the variable in the ODE, as a list.
    shape : tuple
        The shape of the variable.  If not specified, shape is taken from the ODE targets.
    units : str
        The units of the variable.  If not specified, units are taken from the ODE targets.
    static_target : bool
        True if the target is static, otherwise False.

    Notes
    -----
    This method requires that the ODE has run its setup and configure methods.  Thus,
    this method should be called from configure of some parent Group, and the ODE should
    be a system within that Group.
    """
    ode_inputs = ode if isinstance(ode, dict) else get_promoted_vars(ode, iotypes='input')

    targets = get_targets(ode_inputs, name, user_targets=user_targets, control_rates=control_rate)

    if not targets:
        return targets, user_shape, user_units, False

    for tgt in targets:
        if tgt not in ode_inputs:
            raise ValueError(f"No such ODE input: '{tgt}'.")

    if user_units is _unspecified:
        target_units_set = {ode_inputs[tgt]['units'] for tgt in targets}
        if len(target_units_set) == 1:
            units = next(iter(target_units_set))
        else:
            raise ValueError(f'Unable to automatically assign units to {name}. '
                             f'Targets have multiple units: {target_units_set}. '
                             f'Either promote targets and use set_input_defaults to assign common '
                             f'units, or explicitly provide them to {name}.')
    else:
        units = user_units

    # Resolve whether the targets is static or dynamic
    static_target_tags = [tgt for tgt in targets if 'dymos.static_target' in ode_inputs[tgt]['tags']]
    if static_target_tags:
        static_target = True
        if not user_static_target:
            raise ValueError(f"User has specified 'static_target = False' for parameter {name},"
                             f"but one or more targets is tagged with "
                             f"'dymos.static_target': {' '.join(static_target_tags)}")
    else:
        if user_static_target is _unspecified:
            static_target = False
        else:
            static_target = user_static_target

    if user_shape in {None, _unspecified}:
        # Resolve target shape
        target_shape_set = {ode_inputs[tgt]['shape'] for tgt in targets}
        if len(target_shape_set) == 1:
            shape = next(iter(target_shape_set))
            if not static_target:
                if len(shape) == 1:
                    shape = (1,)
                else:
                    shape = shape[1:]
        elif len(target_shape_set) == 0:
            raise ValueError(f'Unable to automatically assign a shape to {name}.\n'
                             'Targets for this variable either do not exist or have no shape set.\n'
                             'The shape for this variable must be set explicitly via the '
                             '`shape=<tuple>` argument.')
        else:
            raise ValueError(f'Unable to automatically assign a shape to {name} based on targets. '
                             f'Targets have multiple shapes assigned: {target_shape_set}. '
                             f'Change targets such that all have common shapes.')
    else:
        shape = user_shape

    return targets, shape, units, static_target


def get_source_metadata(ode, src, user_units=_unspecified, user_shape=_unspecified):
    """
    Return the units and shape of output src in the given ODE.

    If the targets of the state is _unspecified, and the state name is a top level input name
    in the ODE, then the state values are automatically connected to that top-level input.
    If _unspecified and not a top-level input of the ODE, no connection is made.
    If targets is explicitly None, then no connection is made.
    Otherwise, if the user specified some other string or sequence of strings as targets, then
    those are returned.

    Parameters
    ----------
    ode : openmdao.core.System or dict
        The OpenMDAO system which serves as the ODE for dymos, or a dictionary of promoted output names and
        corresponding metadata from get_promoted_vars.  If given as a system, it should already have
        had its setup and configure methods called.
    src : str
        The relative path in the ODE to the source variable whose metadata is requested.
    user_units : str or None or Sequence or _unspecified
        Units for the variable as given by the user.
    user_shape : str or None or Sequence or _unspecified
        Shape for the variable as given by the user.

    Returns
    -------
    dict
        A dictionary containing the metadata for the source. This consists of shape, units, and tags.

    Notes
    -----
    This method requires that the ODE has run its setup and configure methods.  Thus,
    this method should be called from configure of some parent Group, and the ODE should
    be a system within that Group.
    """
    meta = {}
    ode_outputs = ode if isinstance(ode, dict) else get_promoted_vars(ode, iotypes='output')

    if src not in ode_outputs:
        raise ValueError(f'Unable to find the source {src} in the ODE.')

    if user_units in {None, _unspecified}:
        meta['units'] = ode_outputs[src]['units']
    else:
        meta['units'] = user_units

    if user_shape in {None, _unspecified}:
        ode_shape = ode_outputs[src]['shape']
        meta['shape'] = (1,) if len(ode_shape) == 1 else ode_shape[1:]
    else:
        meta['shape'] = user_shape

    meta['tags'] = ode_outputs[src]['tags']

    return meta



import unittest

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

from dymos.examples.finite_burn_orbit_raise.finite_burn_orbit_raise_problem import two_burn_orbit_raise_problem
from dymos.utils.testing_utils import assert_cases_equal


# This test is separate because connected phases aren't directly parallelizable.
@require_pyoptsparse(optimizer='IPOPT')
@use_tempdirs
class TestExampleTwoBurnOrbitRaiseConnectedRestart(unittest.TestCase):

    def test_ex_two_burn_orbit_raise_connected(self):
        optimizer = 'IPOPT'

        p = two_burn_orbit_raise_problem(transcription='gauss-lobatto', transcription_order=3,
                                         compressed=False, optimizer=optimizer,
                                         show_output=False, connected=True)

        if p.model.traj.phases.burn2 in p.model.traj.phases._subsystems_myproc:
            assert_near_equal(p.get_val('traj.burn2.states:deltav')[0], 0.3995,
                              tolerance=4.0E-3)

        case1 = om.CaseReader('dymos_solution.db').get_case('final')
        sim_case1 = om.CaseReader('dymos_simulation.db').get_case('final')

        # Run again without an actual optimizer
        p = two_burn_orbit_raise_problem(transcription='gauss-lobatto', transcription_order=3,
                                         compressed=False, optimizer=optimizer,
                                         show_output=False, connected=True, run_driver=False,
                                         restart='dymos_solution.db')

        p.run_model()

        sim_case2 = om.CaseReader('dymos_simulation.db').get_case('final')

        # Verify that the second case has the same inputs and outputs
        assert_cases_equal(case1, p, tol=1.0E-8)
        assert_cases_equal(sim_case1, sim_case2, tol=1.0E-8)

    def test_restart_from_solution_radau(self):
        optimizer = 'IPOPT'

        p = two_burn_orbit_raise_problem(transcription='radau', transcription_order=3,
                                         compressed=False, optimizer=optimizer, show_output=False)

        case1 = om.CaseReader('dymos_solution.db').get_case('final')
        sim_case1 = om.CaseReader('dymos_simulation.db').get_case('final')

        if p.model.traj.phases.burn2 in p.model.traj.phases._subsystems_myproc:
            assert_near_equal(p.get_val('traj.burn2.states:deltav')[-1], 0.3995,
                              tolerance=2.0E-3)

        # Run again without an actual optimzier
        two_burn_orbit_raise_problem(transcription='radau', transcription_order=3,
                                     compressed=False, optimizer=optimizer, run_driver=False,
                                     show_output=False, restart='dymos_solution.db')

        sim_case2 = om.CaseReader('dymos_simulation.db').get_case('final')

        # Verify that the second case has the same inputs and outputs
        assert_cases_equal(case1, p, tol=1.0E-9)
        assert_cases_equal(sim_case1, sim_case2, tol=1.0E-8)


# This test is separate because connected phases aren't directly parallelizable.
@require_pyoptsparse(optimizer='IPOPT')
@use_tempdirs
class TestExampleTwoBurnOrbitRaiseConnected(unittest.TestCase):

    def test_ex_two_burn_orbit_raise_connected(self):
        optimizer = 'IPOPT'

        p = two_burn_orbit_raise_problem(transcription='gauss-lobatto', transcription_order=3,
                                         compressed=False, optimizer=optimizer,
                                         show_output=False, connected=True)

        if p.model.traj.phases.burn2 in p.model.traj.phases._subsystems_myproc:
            assert_near_equal(p.get_val('traj.burn2.states:deltav')[0], 0.3995,
                              tolerance=4.0E-3)

        case1 = om.CaseReader('dymos_solution.db').get_case('final')
        sim_case1 = om.CaseReader('dymos_simulation.db').get_case('final')

        # Run again without an actual optimizer
        p = two_burn_orbit_raise_problem(transcription='gauss-lobatto', transcription_order=3,
                                         compressed=False, optimizer=optimizer, run_driver=False,
                                         show_output=False, restart='dymos_solution.db',
                                         connected=True)

        sim_case2 = om.CaseReader('dymos_simulation.db').get_case('final')

        # Verify that the second case has the same inputs and outputs
        assert_cases_equal(case1, p, tol=1.0E-8)
        assert_cases_equal(sim_case1, sim_case2, tol=1.0E-8)

    def test_restart_from_solution_radau_to_connected(self):
        optimizer = 'IPOPT'

        p = two_burn_orbit_raise_problem(transcription='radau', transcription_order=3,
                                         compressed=False, optimizer=optimizer, show_output=False)

        case1 = om.CaseReader('dymos_solution.db').get_case('final')
        sim_case1 = om.CaseReader('dymos_simulation.db').get_case('final')

        if p.model.traj.phases.burn2 in p.model.traj.phases._subsystems_myproc:
            assert_near_equal(p.get_val('traj.burn2.states:deltav')[-1], 0.3995,
                              tolerance=2.0E-3)

        # Run again without an actual optimzier
        two_burn_orbit_raise_problem(transcription='radau', transcription_order=3,
                                     compressed=False, optimizer=optimizer, run_driver=False,
                                     show_output=False, restart='dymos_solution.db', connected=True)

        sim_case2 = om.CaseReader('dymos_simulation.db').get_case('final')

        # Verify that the second case has the same inputs and outputs
        assert_cases_equal(case1, p, tol=1.0E-9, require_same_vars=False)
        assert_cases_equal(sim_case1, sim_case2, tol=1.0E-8)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

import dymos as dm
from dymos.utils.lgl import lgl
from dymos.examples.aircraft_steady_flight.aircraft_ode import AircraftODE


@require_pyoptsparse(optimizer='SLSQP')
def ex_aircraft_steady_flight(optimizer='SLSQP', solve_segments=False,
                              use_boundary_constraints=False, compressed=False):
    p = om.Problem(model=om.Group())
    p.driver = om.pyOptSparseDriver()
    p.driver.options['optimizer'] = optimizer
    p.driver.declare_coloring()
    if optimizer == 'SNOPT':
        p.driver.opt_settings['Major iterations limit'] = 20
        p.driver.opt_settings['Major feasibility tolerance'] = 1.0E-6
        p.driver.opt_settings['Major optimality tolerance'] = 1.0E-6
        p.driver.opt_settings["Linesearch tolerance"] = 0.10
        p.driver.opt_settings['iSumm'] = 6
    if optimizer == 'SLSQP':
        p.driver.opt_settings['MAXIT'] = 50

    num_seg = 15
    seg_ends, _ = lgl(num_seg + 1)

    phase = dm.Phase(ode_class=AircraftODE,
                     transcription=dm.Radau(num_segments=num_seg, segment_ends=seg_ends,
                                            order=3, compressed=compressed,
                                            solve_segments=solve_segments))

    # Pass Reference Area from an external source
    assumptions = p.model.add_subsystem('assumptions', om.IndepVarComp())
    assumptions.add_output('S', val=427.8, units='m**2')
    assumptions.add_output('mass_empty', val=1.0, units='kg')
    assumptions.add_output('mass_payload', val=1.0, units='kg')

    p.model.add_subsystem('phase0', phase)

    phase.set_time_options(fix_initial=True,
                           duration_bounds=(300, 10000),
                           duration_ref=5600)

    fix_final = True
    if use_boundary_constraints:
        fix_final = False
        phase.add_boundary_constraint('mass_fuel', loc='final',
                                      equals=1e-3, linear=False)
        phase.add_boundary_constraint('alt', loc='final', equals=10.0, linear=False)

    phase.add_state('range', units='NM',
                    rate_source='range_rate_comp.dXdt:range',
                    fix_initial=True, fix_final=False, ref=1e-3,
                    defect_ref=1e-3, lower=0, upper=2000)

    phase.add_state('mass_fuel', units='lbm',
                    rate_source='propulsion.dXdt:mass_fuel',
                    fix_initial=True, fix_final=fix_final,
                    upper=1.5E5, lower=0.0, ref=1e2, defect_ref=1e2)

    phase.add_state('alt', units='kft',
                    rate_source='climb_rate',
                    fix_initial=True, fix_final=fix_final,
                    lower=0.0, upper=60, ref=1e-3, defect_ref=1e-3)

    phase.add_control('climb_rate', units='ft/min', opt=True, lower=-3000, upper=3000,
                      targets=['gam_comp.climb_rate'],
                      rate_continuity=True, rate2_continuity=False)

    phase.add_control('mach', targets=['tas_comp.mach', 'aero.mach'], opt=False)

    phase.add_parameter('S',
                        targets=['aero.S', 'flight_equilibrium.S', 'propulsion.S'],
                        units='m**2')

    phase.add_parameter('mass_empty', targets=['mass_comp.mass_empty'], units='kg')
    phase.add_parameter('mass_payload', targets=['mass_comp.mass_payload'], units='kg')

    phase.add_path_constraint('propulsion.tau', lower=0.01, upper=2.0)

    p.model.connect('assumptions.S', 'phase0.parameters:S')
    p.model.connect('assumptions.mass_empty', 'phase0.parameters:mass_empty')
    p.model.connect('assumptions.mass_payload', 'phase0.parameters:mass_payload')

    phase.add_objective('range', loc='final', ref=-1.0e-4)

    p.setup()

    p['phase0.t_initial'] = 0.0
    p['phase0.t_duration'] = 3600.0
    p['phase0.states:range'] = phase.interp('range', (0, 724.0))
    p['phase0.states:mass_fuel'] = phase.interp('mass_fuel', (30000, 1e-3))
    p['phase0.states:alt'][:] = 10.0

    p['phase0.controls:mach'][:] = 0.8

    p['assumptions.S'] = 427.8
    p['assumptions.mass_empty'] = 0.15E6
    p['assumptions.mass_payload'] = 84.02869 * 400

    dm.run_problem(p)

    return p


@use_tempdirs
class TestExSteadyAircraftFlight(unittest.TestCase):

    def test_ex_aircraft_steady_flight_opt(self):
        p = ex_aircraft_steady_flight(optimizer='SLSQP', solve_segments=False)
        assert_near_equal(p.get_val('phase0.timeseries.states:range', units='NM')[-1],
                          726.85, tolerance=1.0E-2)

    def test_ex_aircraft_steady_flight_solve(self):
        p = ex_aircraft_steady_flight(optimizer='SLSQP', solve_segments='forward',
                                      use_boundary_constraints=True)
        assert_near_equal(p.get_val('phase0.timeseries.states:range', units='NM')[-1],
                          726.85, tolerance=1.0E-2)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from dymos.utils.testing_utils import assert_check_partials
from dymos.examples.min_time_climb.aero.kappa_comp import KappaComp


class TestKappaComp(unittest.TestCase):

    def test_value(self):
        n = 500
        p = om.Problem()
        p.model.add_subsystem(name='kappa_comp', subsys=KappaComp(num_nodes=n))
        p.setup()
        p.set_val('kappa_comp.mach', np.linspace(0, 1.8, n))
        p.run_model()

        M = p.get_val('kappa_comp.mach')
        kappa = p.get_val('kappa_comp.kappa')

        idxs_0 = np.where(M <= 1.15)[0]
        idxs_1 = np.where(M > 1.15)[0]

        kappa_analtic_0 = 0.54 + 0.15 * (1.0 + np.tanh((M[idxs_0] - 0.9)/0.06))
        kappa_analtic_1 = 0.54 + 0.15 * (1.0 + np.tanh(0.25/0.06)) + 0.14 * (M[idxs_1] - 1.15)

        assert_near_equal(kappa[idxs_0], kappa_analtic_0)
        assert_near_equal(kappa[idxs_1], kappa_analtic_1)

    def test_partials(self):
        n = 10
        p = om.Problem(model=om.Group())
        p.model.add_subsystem(name='kappa_comp', subsys=KappaComp(num_nodes=n))
        p.setup()
        p.set_val('kappa_comp.mach', np.linspace(0, 1.8, n))
        p.run_model()
        cpd = p.check_partials(compact_print=False, out_stream=None)
        assert_check_partials(cpd, atol=1.0E-5, rtol=1.0E-4)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import numpy as np


def split_segments(old_seg_ends, B):
    """
    Compute the new segment ends for the refined grid by splitting the necessary segments.

    Parameters
    ----------
    old_seg_ends : ndarray
        Segment ends of grid on which the problem was solved.
    B : ndarray of ints
        Number of segments to be split into.

    Returns
    -------
    ndarray
        Segment ends of refined grid.
    """
    new_segment_ends = []
    for q in range(0, B.size):
        new_ends = np.linspace(old_seg_ends[q], old_seg_ends[q + 1], B[q] + 1)
        new_segment_ends.extend(new_ends[:-1])
    new_segment_ends.append(1)
    return np.asarray(new_segment_ends)


class PHAdaptive:
    """
    Grid refinement object for the p-then-h grid refinement algorithm.

    The error on a solved phase is evaluated. If error exceeds chosen tolerance, the grid is refined following the
    p-then-h refinement algorithm.
    Patterson, M. A., Hager, W. W., and Rao. A. V., “A ph Mesh Refinement Method for Optimal Control”,
    Optimal Control Applications and Methods, Vol. 36, No. 4, July - August 2015, pp. 398 - 421. DOI: 10.1002/oca2114

    Parameters
    ----------
    phases : Phase
        The Phase object representing the solved phase.
    """

    def __init__(self, phases):
        self.phases = phases
        self.error = {}

    def refine(self, refine_results, iter_number):
        """
        Refine the grid.

        Compute the order, number of nodes, and segment ends required for the new grid
        and assigns them to the transcription of each phase.

        Parameters
        ----------
        refine_results : dict
            A dictionary where each key is the path to a phase in the problem, and the
            associated value are various properties of that phase needed by the refinement
            algorithm.  refine_results is returned by check_error.  This method modifies it
            in place, adding the new_num_segments, new_order, and new_segment_ends.
        iter_number : int
            Iteration number of the grid refinement. The first iteration is 1.

        Returns
        -------
        dict
            A dictionary of phase paths : phases which were refined.
        """
        for phase_path, phase_refinement_results in refine_results.items():
            phase = self.phases[phase_path]
            tx = phase.options['transcription']
            gd = tx.grid_data

            need_refine = phase_refinement_results['need_refinement']
            if not phase.refine_options['refine'] or not np.any(need_refine):
                refine_results[phase_path]['new_order'] = gd.transcription_order
                refine_results[phase_path]['new_num_segments'] = gd.num_segments
                refine_results[phase_path]['new_segment_ends'] = gd.segment_ends
                continue

            # Refinement is needed
            gd = phase.options['transcription'].grid_data
            numseg = gd.num_segments

            refine_seg_idxs = np.where(need_refine)
            P = np.zeros(numseg)

            max_rel_error = refine_results[phase_path]['max_rel_error'][refine_seg_idxs]
            tol = phase.refine_options['tolerance']
            order = gd.transcription_order[refine_seg_idxs]

            P[refine_seg_idxs] = np.log(max_rel_error / tol) / np.log(order)
            P = np.ceil(P).astype(int)

            if gd.transcription == 'gauss-lobatto':
                odd_idxs = np.where(P % 2 != 0)
                P[odd_idxs] += 1

            new_order = gd.transcription_order + P
            B = np.ones(numseg, dtype=int)

            raise_order_idxs = np.where(gd.transcription_order + P <= phase.refine_options['max_order'])
            split_seg_idxs = np.where(gd.transcription_order + P > phase.refine_options['max_order'])

            new_order[raise_order_idxs] = gd.transcription_order[raise_order_idxs] + P[raise_order_idxs]
            new_order[split_seg_idxs] = phase.refine_options['min_order']

            B[split_seg_idxs] = np.around((gd.transcription_order[split_seg_idxs] +
                                           P[split_seg_idxs]) / phase.refine_options['min_order']).astype(int)

            new_order = np.repeat(new_order, repeats=B)
            new_num_segments = int(np.sum(B))
            new_segment_ends = split_segments(gd.segment_ends, B)

            refine_results[phase_path]['new_order'] = new_order
            refine_results[phase_path]['new_num_segments'] = new_num_segments
            refine_results[phase_path]['new_segment_ends'] = new_segment_ends

            tx.options['order'] = new_order
            tx.options['num_segments'] = new_num_segments
            tx.options['segment_ends'] = new_segment_ends
            tx.init_grid()



import numpy as np
from ....utils.misc import get_rate_units
from ....options import options as dymos_options
import openmdao.api as om


class StateRateCollectorComp(om.ExplicitComponent):
    """
    Class definition for StateRateCollectorComp.

    Collects the state rates and outputs them in the units specified in the state options.
    For explicit integration this is necessary when the output providing the state rate has
    different units than those defined in the state_options/time_options.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def initialize(self):
        """
        Declare component options.
        """
        self.options.declare(
            'state_options', types=dict,
            desc='Dictionary of options for the ODE state variables.')
        self.options.declare(
            'time_units', default=None, allow_none=True, types=str,
            desc='Units of time')

        self._no_check_partials = not dymos_options['include_check_partials']

    def setup(self):
        """
        Create inputs/outputs on this component.
        """
        state_options = self.options['state_options']
        time_units = self.options['time_units']

        for name, options in state_options.items():
            shape = options['shape']
            units = options['units']

            rate_units = get_rate_units(units, time_units)

            self.add_input(f'state_rates_in:{name}_rate', val=np.ones(shape), units=rate_units)
            self.add_output( f'state_rates:{name}_rate', shape=shape, units=rate_units)

    def compute(self, inputs, outputs):
        """
        Compute component outputs.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        """
        outputs.set_val(inputs.asarray())



import openmdao.api as om
import numpy as np


class TireConstraintODE(om.ExplicitComponent):
    def initialize(self):
        self.options.declare("num_nodes", types=int)

    def setup(self):
        nn = self.options["num_nodes"]

        # constants
        self.add_input("mu0_y", val=1.68, desc="lateral friction coefficient", units=None)
        self.add_input("mu0_x", val=1.68, desc="longitudinal friction coefficient", units=None)
        self.add_input("k_mu", val=-0.0, desc="tire load sensitivity", units=None)
        self.add_input("a", val=1.8, desc="cg to front distance", units="m")
        self.add_input("b", val=1.6, desc="cg to rear distance", units="m")
        self.add_input("M", val=0.0, desc="mass", units="kg")
        self.add_input("g", val=9.8, desc="mass", units="m/s**2")

        # states
        self.add_input("S_fl", val=np.zeros(nn), desc="longitudinal force fl", units="N")
        self.add_input("S_fr", val=np.zeros(nn), desc="longitudinal force fr", units="N")
        self.add_input("S_rl", val=np.zeros(nn), desc="longitudinal force rl", units="N")
        self.add_input("S_rr", val=np.zeros(nn), desc="longitudinal force rr", units="N")

        self.add_input("F_fl", val=np.zeros(nn), desc="lateral force fl", units="N")
        self.add_input("F_fr", val=np.zeros(nn), desc="lateral force fr", units="N")
        self.add_input("F_rl", val=np.zeros(nn), desc="lateral force rl", units="N")
        self.add_input("F_rr", val=np.zeros(nn), desc="lateral force rr", units="N")

        self.add_input("N_rr", val=np.zeros(nn), desc="normal load rr", units="N")
        self.add_input("N_fr", val=np.zeros(nn), desc="normal load fr", units="N")
        self.add_input("N_rl", val=np.zeros(nn), desc="normal load rl", units="N")
        self.add_input("N_fl", val=np.zeros(nn), desc="normal load fl", units="N")

        # outputs
        self.add_output("c_rr", val=np.zeros(nn), desc="tire load constraint rr", units=None)
        self.add_output("c_rl", val=np.zeros(nn), desc="tire load constraint rr", units=None)
        self.add_output("c_fr", val=np.zeros(nn), desc="tire load constraint rr", units=None)
        self.add_output("c_fl", val=np.zeros(nn), desc="tire load constraint rr", units=None)

        # # Setup partials
        arange = np.arange(self.options["num_nodes"], dtype=int)

        # #partials
        self.declare_partials(of="c_rr", wrt="S_rr", rows=arange, cols=arange)
        self.declare_partials(of="c_rr", wrt="F_rr", rows=arange, cols=arange)
        self.declare_partials(of="c_rr", wrt="N_rr", rows=arange, cols=arange)

        self.declare_partials(of="c_rl", wrt="S_rl", rows=arange, cols=arange)
        self.declare_partials(of="c_rl", wrt="F_rl", rows=arange, cols=arange)
        self.declare_partials(of="c_rl", wrt="N_rl", rows=arange, cols=arange)

        self.declare_partials(of="c_fr", wrt="S_fr", rows=arange, cols=arange)
        self.declare_partials(of="c_fr", wrt="F_fr", rows=arange, cols=arange)
        self.declare_partials(of="c_fr", wrt="N_fr", rows=arange, cols=arange)

        self.declare_partials(of="c_fl", wrt="S_fl", rows=arange, cols=arange)
        self.declare_partials(of="c_fl", wrt="F_fl", rows=arange, cols=arange)
        self.declare_partials(of="c_fl", wrt="N_fl", rows=arange, cols=arange)

    def compute(self, inputs, outputs):
        mu0_y = inputs["mu0_y"]
        mu0_x = inputs["mu0_x"]
        k_mu = inputs["k_mu"]
        a = inputs["a"]
        b = inputs["b"]
        M = inputs["M"]
        g = inputs["g"]

        S_fl = inputs["S_fl"]
        S_fr = inputs["S_fr"]
        S_rl = inputs["S_rl"]
        S_rr = inputs["S_rr"]

        N_fl = inputs["N_fl"]
        N_fr = inputs["N_fr"]
        N_rl = inputs["N_rl"]
        N_rr = inputs["N_rr"]

        F_fl = inputs["F_fl"]
        F_fr = inputs["F_fr"]
        F_rl = inputs["F_rl"]
        F_rr = inputs["F_rr"]

        N0_rr = (M * g / 2) * (a / (a + b))
        N0_rl = (M * g / 2) * (a / (a + b))
        N0_fr = (M * g / 2) * (b / (a + b))
        N0_fl = (M * g / 2) * (b / (a + b))

        mu_x_rr = mu0_x + k_mu * N_rr / N0_rr
        mu_x_rl = mu0_x + k_mu * N_rl / N0_rl
        mu_x_fr = mu0_x + k_mu * N_fr / N0_fr
        mu_x_fl = mu0_x + k_mu * N_fl / N0_fl

        mu_y_rr = mu0_y + k_mu * N_rr / N0_rr
        mu_y_rl = mu0_y + k_mu * N_rl / N0_rl
        mu_y_fr = mu0_y + k_mu * N_fr / N0_fr
        mu_y_fl = mu0_y + k_mu * N_fl / N0_fl

        outputs["c_rr"] = (S_rr / (N_rr * mu_x_rr)) ** 2 + (F_rr / (N_rr * mu_y_rr)) ** 2
        outputs["c_rl"] = (S_rl / (N_rl * mu_x_rl)) ** 2 + (F_rl / (N_rl * mu_y_rl)) ** 2
        outputs["c_fr"] = (S_fr / (N_fr * mu_x_fr)) ** 2 + (F_fr / (N_fr * mu_y_fr)) ** 2
        outputs["c_fl"] = (S_fl / (N_fl * mu_x_fl)) ** 2 + (F_fl / (N_fl * mu_y_fl)) ** 2

    def compute_partials(self, inputs, jacobian):
        mu0_y = inputs["mu0_y"]
        mu0_x = inputs["mu0_x"]
        k_mu = inputs["k_mu"]
        a = inputs["a"]
        b = inputs["b"]
        M = inputs["M"]
        g = inputs["g"]

        S_fl = inputs["S_fl"]
        S_fr = inputs["S_fr"]
        S_rl = inputs["S_rl"]
        S_rr = inputs["S_rr"]

        N_fl = inputs["N_fl"]
        N_fr = inputs["N_fr"]
        N_rl = inputs["N_rl"]
        N_rr = inputs["N_rr"]

        F_fl = inputs["F_fl"]
        F_fr = inputs["F_fr"]
        F_rl = inputs["F_rl"]
        F_rr = inputs["F_rr"]

        N0_rr = (M * g / 2) * (a / (a + b))
        N0_rl = (M * g / 2) * (a / (a + b))
        N0_fr = (M * g / 2) * (b / (a + b))
        N0_fl = (M * g / 2) * (b / (a + b))

        jacobian["c_rr", "S_rr"] = (2 * S_rr) / (N_rr ** 2 * (mu0_x + (N_rr * k_mu) / N0_rr) ** 2)
        jacobian["c_fr", "S_fr"] = (2 * S_fr) / (N_fr ** 2 * (mu0_x + (N_fr * k_mu) / N0_fr) ** 2)
        jacobian["c_fl", "S_fl"] = (2 * S_fl) / (N_fl ** 2 * (mu0_x + (N_fl * k_mu) / N0_fl) ** 2)
        jacobian["c_rl", "S_rl"] = (2 * S_rl) / (N_rl ** 2 * (mu0_x + (N_rl * k_mu) / N0_rl) ** 2)

        jacobian["c_rr", "F_rr"] = (2 * F_rr) / (N_rr ** 2 * (mu0_y + (N_rr * k_mu) / N0_rr) ** 2)
        jacobian["c_fr", "F_fr"] = (2 * F_fr) / (N_fr ** 2 * (mu0_y + (N_fr * k_mu) / N0_fr) ** 2)
        jacobian["c_fl", "F_fl"] = (2 * F_fl) / (N_fl ** 2 * (mu0_y + (N_fl * k_mu) / N0_fl) ** 2)
        jacobian["c_rl", "F_rl"] = (2 * F_rl) / (N_rl ** 2 * (mu0_y + (N_rl * k_mu) / N0_rl) ** 2)

        jacobian["c_rr", "N_rr"] = (
            -(2 * F_rr ** 2) / (N_rr ** 3 * (mu0_y + (k_mu * N_rr) / N0_rr) ** 2) -
            (2 * S_rr ** 2) / (N_rr ** 3 * (mu0_x + (k_mu * N_rr) / N0_rr) ** 2) -
            (2 * F_rr ** 2 * k_mu) /
            (N_rr ** 2 * N0_rr * (mu0_y + (k_mu * N_rr) / N0_rr) ** 3) -
            (2 * k_mu * S_rr ** 2) /
            (N_rr ** 2 * N0_rr * (mu0_x + (k_mu * N_rr) / N0_rr) ** 3)
        )
        jacobian["c_rl", "N_rl"] = (
            -(2 * F_rl ** 2) / (N_rl ** 3 * (mu0_y + (k_mu * N_rl) / N0_rl) ** 2) -
            (2 * S_rl ** 2) / (N_rl ** 3 * (mu0_x + (k_mu * N_rl) / N0_rl) ** 2) -
            (2 * F_rl ** 2 * k_mu) /
            (N_rl ** 2 * N0_rl * (mu0_y + (k_mu * N_rl) / N0_rl) ** 3) -
            (2 * k_mu * S_rl ** 2) /
            (N_rl ** 2 * N0_rl * (mu0_x + (k_mu * N_rl) / N0_rl) ** 3)
        )
        jacobian["c_fr", "N_fr"] = (
            -(2 * F_fr ** 2) / (N_fr ** 3 * (mu0_y + (k_mu * N_fr) / N0_fr) ** 2) -
            (2 * S_fr ** 2) / (N_fr ** 3 * (mu0_x + (k_mu * N_fr) / N0_fr) ** 2) -
            (2 * F_fr ** 2 * k_mu) /
            (N_fr ** 2 * N0_fr * (mu0_y + (k_mu * N_fr) / N0_fr) ** 3) -
            (2 * k_mu * S_fr ** 2) /
            (N_fr ** 2 * N0_fr * (mu0_x + (k_mu * N_fr) / N0_fr) ** 3)
        )
        jacobian["c_fl", "N_fl"] = (
            -(2 * F_fl ** 2) / (N_fl ** 3 * (mu0_y + (k_mu * N_fl) / N0_fl) ** 2) -
            (2 * S_fl ** 2) / (N_fl ** 3 * (mu0_x + (k_mu * N_fl) / N0_fl) ** 2) -
            (2 * F_fl ** 2 * k_mu) /
            (N_fl ** 2 * N0_fl * (mu0_y + (k_mu * N_fl) / N0_fl) ** 3) -
            (2 * k_mu * S_fl ** 2) /
            (N_fl ** 2 * N0_fl * (mu0_x + (k_mu * N_fl) / N0_fl) ** 3)
        )



import matplotlib.pyplot as plt


def plot_results(axes, title, figsize=(10, 8), p_sol=None, p_sim=None):
    """
    Plot the timeseries results of a Dymos problem using matplotlib.

    Parameters
    ----------
    axes : list of tuple of (str, str, str, str)
        A sequence of tuples wherein each tuple contains ('x path', 'y path' , 'x label', 'y label').
    title : str
        The figure title
    figsize : tuple of (int, int)
        The size of the created figure, in inches
    p_sol : The solution problem instance.
    p_sim : The simulation problem instance.

    Returns
    -------
    fig, axes
        The Figure object and sequence of axes associated with the plot.

    """
    nrows = len(axes)

    fig, axs = plt.subplots(nrows=nrows, ncols=1, figsize=figsize)
    fig.suptitle(title)

    if nrows == 1:
        axs = [axs]

    for i, (x, y, xlabel, ylabel) in enumerate(axes):
        if p_sol is not None:
            axs[i].plot(p_sol.get_val(x),
                        p_sol.get_val(y),
                        marker='o',
                        ms=4,
                        linestyle='None',
                        label='solution' if i == 0 else None)

        if p_sim is not None:
            axs[i].plot(p_sim.get_val(x),
                        p_sim.get_val(y),
                        marker=None,
                        linestyle='-',
                        label='simulation' if i == 0 else None)

        axs[i].set_xlabel(xlabel)
        axs[i].set_ylabel(ylabel)
        fig.suptitle(title)
        fig.legend(loc='lower center', ncol=2)

    return fig, axs



import itertools
import os
import unittest

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

import dymos as dm
from dymos.examples.double_integrator.double_integrator_ode import DoubleIntegratorODE


@require_pyoptsparse(optimizer='SLSQP')
def double_integrator_direct_collocation(transcription='gauss-lobatto', compressed=True):

    p = om.Problem(model=om.Group())
    p.driver = om.pyOptSparseDriver()
    p.driver.declare_coloring()

    if transcription == 'gauss-lobatto':
        t = dm.GaussLobatto(num_segments=30, order=3, compressed=compressed)
    elif transcription == "radau-ps":
        t = dm.Radau(num_segments=30, order=3, compressed=compressed)
    else:
        raise ValueError('invalid transcription')

    traj = p.model.add_subsystem('traj', dm.Trajectory())

    phase = traj.add_phase('phase0', dm.Phase(ode_class=DoubleIntegratorODE, transcription=t))

    phase.set_time_options(fix_initial=True, fix_duration=True, units='s')

    phase.add_state('v', fix_initial=True, fix_final=True, rate_source='u', units='m/s')
    phase.add_state('x', fix_initial=True, rate_source='v', units='m', shape=(1, ))

    phase.add_control('u', units='m/s**2', scaler=0.01, continuity=False, rate_continuity=False,
                      rate2_continuity=False, shape=(1, ), lower=-1.0, upper=1.0)

    # Maximize distance travelled in one second.
    phase.add_objective('x', loc='final', scaler=-1)

    p.model.linear_solver = om.DirectSolver()

    p.setup(check=True)

    p['traj.phase0.t_initial'] = 0.0
    p['traj.phase0.t_duration'] = 1.0

    p['traj.phase0.states:x'] = phase.interp('x', [0, 0.25])
    p['traj.phase0.states:v'] = phase.interp('v', [0, 0])
    p['traj.phase0.controls:u'] = phase.interp('u', [1, -1])

    p.run_driver()

    return p


@use_tempdirs
class TestDoubleIntegratorExample(unittest.TestCase):

    @classmethod
    def tearDownClass(cls):
        for filename in ['total_coloring.pkl', 'SLSQP.out', 'SNOPT_print.out']:
            if os.path.exists(filename):
                os.remove(filename)

    def _assert_results(self, p, traj=True, tol=1.0E-4):
        if traj:
            x = p.get_val('traj.phase0.timeseries.states:x')
            v = p.get_val('traj.phase0.timeseries.states:v')
        else:
            x = p.get_val('phase0.timeseries.states:x')
            v = p.get_val('phase0.timeseries.states:v')

        assert_near_equal(x[0], 0.0, tolerance=tol)
        assert_near_equal(x[-1], 0.25, tolerance=tol)

        assert_near_equal(v[0], 0.0, tolerance=tol)
        assert_near_equal(v[-1], 0.0, tolerance=tol)

    def test_ex_double_integrator_gl_compressed(self):
        p = double_integrator_direct_collocation('gauss-lobatto',
                                                 compressed=True)

        x = p.get_val('traj.phase0.timeseries.states:x')
        v = p.get_val('traj.phase0.timeseries.states:v')

        assert_near_equal(x[0], 0.0, tolerance=1.0E-4)
        assert_near_equal(x[-1], 0.25, tolerance=1.0E-4)

        assert_near_equal(v[0], 0.0, tolerance=1.0E-4)
        assert_near_equal(v[-1], 0.0, tolerance=1.0E-4)

    def test_ex_double_integrator_gl_uncompressed(self):
        p = double_integrator_direct_collocation('gauss-lobatto',
                                                 compressed=False)

        x = p.get_val('traj.phase0.timeseries.states:x')
        v = p.get_val('traj.phase0.timeseries.states:v')

        assert_near_equal(x[0], 0.0, tolerance=1.0E-4)
        assert_near_equal(x[-1], 0.25, tolerance=1.0E-4)

        assert_near_equal(v[0], 0.0, tolerance=1.0E-4)
        assert_near_equal(v[-1], 0.0, tolerance=1.0E-4)

    def test_ex_double_integrator_radau_compressed(self):
        p = double_integrator_direct_collocation('radau-ps',
                                                 compressed=True)

        x = p.get_val('traj.phase0.timeseries.states:x')
        v = p.get_val('traj.phase0.timeseries.states:v')

        assert_near_equal(x[0], 0.0, tolerance=1.0E-4)
        assert_near_equal(x[-1], 0.25, tolerance=1.0E-4)

        assert_near_equal(v[0], 0.0, tolerance=1.0E-4)
        assert_near_equal(v[-1], 0.0, tolerance=1.0E-4)

    def test_ex_double_integrator_radau_uncompressed(self):
        p = double_integrator_direct_collocation('radau-ps',
                                                 compressed=False)

        x = p.get_val('traj.phase0.timeseries.states:x')
        v = p.get_val('traj.phase0.timeseries.states:v')

        assert_near_equal(x[0], 0.0, tolerance=1.0E-4)
        assert_near_equal(x[-1], 0.25, tolerance=1.0E-4)

        assert_near_equal(v[0], 0.0, tolerance=1.0E-4)
        assert_near_equal(v[-1], 0.0, tolerance=1.0E-4)

    @require_pyoptsparse(optimizer='SLSQP')
    def test_ex_double_integrator_input_times_uncompressed(self):
        """
        Tests that externally connected t_initial and t_duration function as expected.
        """
        compressed = False
        p = om.Problem(model=om.Group())
        p.driver = om.pyOptSparseDriver()
        p.driver.declare_coloring()

        times_ivc = p.model.add_subsystem('times_ivc', om.IndepVarComp(),
                                          promotes_outputs=['t0', 'tp'])
        times_ivc.add_output(name='t0', val=0.0, units='s')
        times_ivc.add_output(name='tp', val=1.0, units='s')

        transcription = dm.Radau(num_segments=20, order=3, compressed=compressed)
        phase = dm.Phase(ode_class=DoubleIntegratorODE, transcription=transcription)
        p.model.add_subsystem('phase0', phase)

        p.model.connect('t0', 'phase0.t_initial')
        p.model.connect('tp', 'phase0.t_duration')

        phase.set_time_options(input_initial=True, input_duration=True, units='s')

        phase.add_state('v', fix_initial=True, fix_final=True, rate_source='u', units='m/s')
        phase.add_state('x', fix_initial=True, rate_source='v', units='m')

        phase.add_control('u', units='m/s**2', scaler=0.01, continuity=False, rate_continuity=False,
                          rate2_continuity=False, shape=(1,), lower=-1.0, upper=1.0)

        # Maximize distance travelled in one second.
        phase.add_objective('x', loc='final', scaler=-1)

        p.model.linear_solver = om.DirectSolver()

        p.setup(check=True)

        p['t0'] = 0.0
        p['tp'] = 1.0

        p['phase0.states:x'] = phase.interp('x', [0, 0.25])
        p['phase0.states:v'] = phase.interp('v', [0, 0])
        p['phase0.controls:u'] = phase.interp('u', [1, -1])

        p.run_driver()

        self._assert_results(p, traj=False)
        exp_out = phase.simulate()
        self._assert_results(exp_out, traj=False, tol=1.0E-2)

    @require_pyoptsparse(optimizer='SLSQP')
    def test_ex_double_integrator_input_times_compressed(self):
        """
        Tests that externally connected t_initial and t_duration function as expected.
        """
        compressed = True
        p = om.Problem(model=om.Group())
        p.driver = om.pyOptSparseDriver()
        p.driver.declare_coloring()

        times_ivc = p.model.add_subsystem('times_ivc', om.IndepVarComp(),
                                          promotes_outputs=['t0', 'tp'])
        times_ivc.add_output(name='t0', val=0.0, units='s')
        times_ivc.add_output(name='tp', val=1.0, units='s')

        transcription = dm.Radau(num_segments=20, order=3, compressed=compressed)
        phase = dm.Phase(ode_class=DoubleIntegratorODE, transcription=transcription)
        p.model.add_subsystem('phase0', phase)

        p.model.connect('t0', 'phase0.t_initial')
        p.model.connect('tp', 'phase0.t_duration')

        phase.set_time_options(input_initial=True, input_duration=True, units='s')

        phase.add_state('v', fix_initial=True, fix_final=True, rate_source='u', units='m/s')
        phase.add_state('x', fix_initial=True, rate_source='v', units='m')

        phase.add_control('u', units='m/s**2', scaler=0.01, continuity=False, rate_continuity=False,
                          rate2_continuity=False, shape=(1, ), lower=-1.0, upper=1.0)

        # Maximize distance travelled in one second.
        phase.add_objective('x', loc='final', scaler=-1)

        p.model.linear_solver = om.DirectSolver()

        p.setup(check=True)

        p['t0'] = 0.0
        p['tp'] = 1.0

        p['phase0.states:x'] = phase.interp('x', [0, 0.25])
        p['phase0.states:v'] = phase.interp('v', [0, 0])
        p['phase0.controls:u'] = phase.interp('u', [1, -1])

        p.run_driver()

        assert_near_equal(p.get_val('phase0.timeseries.states:x')[-1, ...],
                          [0.25],
                          tolerance=1.0E-8)


if __name__ == "__main__":

    unittest.main()



import numpy as np
from scipy.linalg import block_diag
import scipy.sparse as sp

from openmdao.utils.units import unit_conversion

from ....utils.lagrange import lagrange_matrices

from ...common.timeseries_output_comp import TimeseriesOutputCompBase


_USE_SPARSE = True


class PseudospectralTimeseriesOutputComp(TimeseriesOutputCompBase):
    """
    Class definition of the PseudospectralTimeseriesOutputComp.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, **kwargs):
        super(PseudospectralTimeseriesOutputComp, self).__init__(**kwargs)

        # Flag to set if no multiplication by the interpolation matrix is necessary
        self._no_interp = False

    def _declare_options(self):
        """
        Declare options before kwargs are processed in the init method.
        """
        super()._declare_options()
        self.options.declare('time_units', default=None, allow_none=True, types=str,
                             desc='Units of time')

    def setup(self):
        """
        Define the independent variables as output variables.
        """
        igd = self.options['input_grid_data']
        ogd = self.options['output_grid_data']
        output_subset = self.options['output_subset']

        if ogd is None:
            ogd = igd

        if ogd is igd and output_subset == 'all':
            self._no_interp = True

        self.input_num_nodes = igd.num_nodes
        self.output_num_nodes = ogd.subset_num_nodes[output_subset]

        # Build the interpolation matrix which maps from the input grid to the output grid.
        # Rather than a single phase-wide interpolating polynomial, map each segment.
        # To do this, find the nodes in the output grid which fall in each segment of the input
        # grid.  Then build a Lagrange interpolating polynomial for that segment
        L_blocks = []
        D_blocks = []
        output_nodes_ptau = ogd.node_ptau[ogd.subset_node_indices[output_subset]]

        for iseg in range(igd.num_segments):
            i1, i2 = igd.segment_indices[iseg]
            iptau_segi = igd.node_ptau[i1:i2]
            istau_segi = igd.node_stau[i1:i2]

            # The indices of the output grid that fall within this segment of the input grid
            if ogd is igd:
                optau_segi = iptau_segi
            else:
                ptau_hi = igd.segment_ends[iseg+1]
                if iseg < igd.num_segments - 1:
                    optau_segi = output_nodes_ptau[output_nodes_ptau <= ptau_hi]
                else:
                    optau_segi = output_nodes_ptau

                # Remove the captured nodes so we don't accidentally include them again
                output_nodes_ptau = output_nodes_ptau[len(optau_segi):]

            # # Now get the output nodes which fall in iseg in iseg's segment tau space.
            ostau_segi = 2.0 * (optau_segi - iptau_segi[0]) / (iptau_segi[-1] - iptau_segi[0]) - 1

            # Create the interpolation matrix and add it to the blocks
            L, D = lagrange_matrices(istau_segi, ostau_segi)
            L_blocks.append(L)
            D_blocks.append(D)

        if _USE_SPARSE:
            self.interpolation_matrix = sp.block_diag(L_blocks, format='csr')
            self.differentiation_matrix = sp.block_diag(D_blocks, format='csr')
        else:
            self.interpolation_matrix = block_diag(*L_blocks)
            self.differentiation_matrix = block_diag(*D_blocks)

        self.add_input('dt_dstau', shape=(self.input_num_nodes,), units=self.options['time_units'])

    def _add_output_configure(self, name, units, shape, desc='', src=None, rate=False):
        """
        Add a single timeseries output.

        Can be called by parent groups in configure.

        Parameters
        ----------
        name : str
            name of the variable in this component's namespace.
        shape : int or tuple or list or None
            Shape of this variable, only required if val is not an array.
            Default is None.
        units : str or None
            Units in which the output variables will be provided to the component during execution.
            Default is None, which means it has no units.
        desc : str
            description of the timeseries output variable.
        src : str
            The src path of the variables input, used to prevent redundant inputs.
        rate : bool
            If True, timeseries output is a rate.

        Returns
        -------
        bool
            True if a new input was added for the output, or False if it reuses an existing input.
        """
        input_num_nodes = self.input_num_nodes
        output_num_nodes = self.output_num_nodes
        added_source = False

        if name in self._vars:
            return False

        if src in self._sources:
            # If we're already pulling the source into this timeseries, use that as the
            # input for this output.
            input_name = self._sources[src]
            input_units = self._units[input_name]
        else:
            input_name = f'input_values:{name}'
            self.add_input(input_name,
                           shape=(input_num_nodes,) + shape,
                           units=units, desc=desc)
            self._sources[src] = input_name
            input_units = self._units[input_name] = units
            added_source = True

        output_name = name
        self.add_output(output_name, shape=(output_num_nodes,) + shape, units=units, desc=desc)

        self._vars[name] = (input_name, output_name, shape, rate)

        size = np.prod(shape)
        jac = np.zeros((output_num_nodes, size, input_num_nodes, size))

        if rate:
            mat = self.differentiation_matrix
        else:
            mat = self.interpolation_matrix

        for i in range(size):
            if _USE_SPARSE:
                jac[:, i, :, i] = mat.toarray()
            else:
                jac[:, i, :, i] = mat

        jac = jac.reshape((output_num_nodes * size, input_num_nodes * size), order='C')

        jac_rows, jac_cols = np.nonzero(jac)

        # There's a chance that the input for this output was pulled from another variable with
        # different units, so account for that with a conversion.
        if input_units is None or units is None:
            self.declare_partials(of=output_name, wrt=input_name,
                                  rows=jac_rows, cols=jac_cols, val=jac[jac_rows, jac_cols])
        else:
            scale, offset = unit_conversion(input_units, units)
            self._conversion_factors[output_name] = scale, offset

            self.declare_partials(of=output_name, wrt=input_name,
                                  rows=jac_rows, cols=jac_cols,
                                  val=scale * jac[jac_rows, jac_cols])

        return added_source

    def compute(self, inputs, outputs):
        """
        Compute component outputs.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        """
        # convert dt_dstau to a column vector
        dt_dstau = inputs['dt_dstau'][:, np.newaxis]

        for (input_name, output_name, _, is_rate) in self._vars.values():
            if self._no_interp:
                interp_vals = inputs[input_name]
            else:
                inp = inputs[input_name]
                if len(inp.shape) > 2:
                    # Dot product always performs the sum product over axis 2.
                    inp = inp.swapaxes(0, 1)
                interp_vals = self.interpolation_matrix.dot(inp)

            if is_rate:
                interp_vals = self.differentiation_matrix.dot(interp_vals) / dt_dstau

            if output_name in self._conversion_factors:
                scale, offset = self._conversion_factors[output_name]
                outputs[output_name] = scale * (interp_vals + offset)
            else:
                outputs[output_name] = interp_vals



import numpy as np
import openmdao.api as om
from dymos.models.atmosphere import USatm1976Comp

from .steady_flight_path_angle_comp import SteadyFlightPathAngleComp
from .dynamic_pressure_comp import DynamicPressureComp
from .flight_equlibrium.steady_flight_equilibrium_group import SteadyFlightEquilibriumGroup
from .propulsion.propulsion_group import PropulsionGroup
from .range_rate_comp import RangeRateComp
from .true_airspeed_comp import TrueAirspeedComp
from .mass_comp import MassComp


class AircraftODE(om.Group):

    def initialize(self):
        self.options.declare('num_nodes', types=int,
                             desc='Number of nodes to be evaluated in the RHS')

    def setup(self):
        nn = self.options['num_nodes']

        self.add_subsystem(name='mass_comp',
                           subsys=MassComp(num_nodes=nn),
                           promotes_inputs=['mass_fuel'])

        self.connect('mass_comp.W_total', 'flight_equilibrium.W_total')

        self.add_subsystem(name='atmos',
                           subsys=USatm1976Comp(num_nodes=nn),
                           promotes_inputs=[('h', 'alt')])

        self.connect('atmos.pres', 'propulsion.pres')
        self.connect('atmos.sos', 'tas_comp.sos')
        self.connect('atmos.rho', 'q_comp.rho')

        self.add_subsystem(name='tas_comp',
                           subsys=TrueAirspeedComp(num_nodes=nn))

        self.connect('tas_comp.TAS',
                     ('gam_comp.TAS', 'q_comp.TAS', 'range_rate_comp.TAS'))

        self.add_subsystem(name='gam_comp',
                           subsys=SteadyFlightPathAngleComp(num_nodes=nn))

        self.connect('gam_comp.gam', ('flight_equilibrium.gam', 'range_rate_comp.gam'))

        self.add_subsystem(name='q_comp',
                           subsys=DynamicPressureComp(num_nodes=nn))

        self.connect('q_comp.q', ('aero.q', 'flight_equilibrium.q', 'propulsion.q'))

        self.add_subsystem(name='flight_equilibrium',
                           subsys=SteadyFlightEquilibriumGroup(num_nodes=nn),
                           promotes_inputs=['aero.*', 'alt'],
                           promotes_outputs=['aero.*'])

        self.connect('flight_equilibrium.CT', 'propulsion.CT')

        self.add_subsystem(name='propulsion', subsys=PropulsionGroup(num_nodes=nn),
                           promotes_inputs=['alt'])

        self.add_subsystem(name='range_rate_comp', subsys=RangeRateComp(num_nodes=nn))

        # We promoted multiple inputs to the group named 'alt'.
        # In order to automatically create a source variable for 'alt', we must specify their units
        # since they have different units in different components.
        self.set_input_defaults('alt', val=np.ones(nn,), units='m')



import numpy as np

import dymos as dm
from dymos.examples.water_rocket.water_propulsion_ode import WaterPropulsionODE


def new_propelled_ascent_phase(transcription):
    propelled_ascent = dm.Phase(ode_class=WaterPropulsionODE,
                                transcription=transcription)

    # All initial states except flight path angle and water volume are fixed
    # Final flight path angle is fixed (we will set it to zero so that the phase ends at apogee)
    # Final water volume is fixed (we will set it to zero so that phase ends when bottle empties)
    propelled_ascent.set_time_options(
        fix_initial=True, duration_bounds=(0.001, 0.5), duration_ref=0.1, units='s')
    propelled_ascent.add_state('r', units='m', rate_source='eom.r_dot',
                               fix_initial=True, fix_final=False, ref=100.0, defect_ref=100.0)
    propelled_ascent.add_state('h', units='m', rate_source='eom.h_dot', targets=['atmos.h'],
                               fix_initial=True, fix_final=False, ref=100.0, defect_ref=100.0)
    propelled_ascent.add_state('gam', units='deg', rate_source='eom.gam_dot', targets=['eom.gam'],
                               fix_initial=False, fix_final=False, lower=0, upper=85.0, ref=90)
    propelled_ascent.add_state('v', units='m/s', rate_source='eom.v_dot', targets=['dynamic_pressure.v', 'eom.v'],
                               fix_initial=True, fix_final=False, ref=100, defect_ref=100, lower=0.01)

    propelled_ascent.add_state('p', units='bar', rate_source='water_engine.pdot',
                               targets=['water_engine.p'], fix_initial=True, fix_final=False,
                               lower=1.02)
    propelled_ascent.add_state('V_w', units='L', rate_source='water_engine.Vdot',
                               targets=['water_engine.V_w', 'mass_adder.V_w'],
                               fix_initial=False, fix_final=True, ref=10, defect_ref=10,
                               lower=0)

    propelled_ascent.add_parameter(
        'S', targets=['aero.S'], units='m**2')
    propelled_ascent.add_parameter(
        'm_empty', targets=['mass_adder.m_empty'], units='kg')
    propelled_ascent.add_parameter(
        'V_b', targets=['water_engine.V_b'], units='m**3')

    propelled_ascent.add_timeseries_output('water_engine.F', 'T', units='N')

    return propelled_ascent


def new_ballistic_ascent_phase(transcription):
    ballistic_ascent = dm.Phase(ode_class=WaterPropulsionODE, transcription=transcription,
                                ode_init_kwargs={'ballistic': True})

    # All initial states are free (they will be  linked to the final stages of propelled_ascent).
    # Final flight path angle is fixed (we will set it to zero so that the phase ends at apogee)
    ballistic_ascent.set_time_options(
        fix_initial=False, initial_bounds=(0.001, 1), duration_bounds=(0.001, 10),
        duration_ref=1, units='s')

    ballistic_ascent.add_state('r', units='m', rate_source='eom.r_dot', fix_initial=False,
                               fix_final=False, ref=100.0, defect_ref=100.0)
    ballistic_ascent.add_state('h', units='m', rate_source='eom.h_dot', targets=['atmos.h'],
                               fix_initial=False, fix_final=False, ref=100.0, defect_ref=100.0)
    ballistic_ascent.add_state('gam', units='deg', rate_source='eom.gam_dot', targets=['eom.gam'],
                               fix_initial=False, fix_final=True, upper=89, ref=90, defect_ref=90)
    ballistic_ascent.add_state('v', units='m/s', rate_source='eom.v_dot',
                               targets=['dynamic_pressure.v', 'eom.v'], fix_initial=False,
                               fix_final=False, ref=100, defect_ref=100)

    ballistic_ascent.add_parameter('S', targets=['aero.S'], units='m**2')
    ballistic_ascent.add_parameter('m_empty', targets=['eom.m'], units='kg')

    return ballistic_ascent


def new_descent_phase(transcription):
    descent = dm.Phase(ode_class=WaterPropulsionODE, transcription=transcription,
                       ode_init_kwargs={'ballistic': True})

    # All initial states and time are free (they will be linked to the final states of ballistic_ascent).
    # Final altitude is fixed (we will set it to zero so that the phase ends at ground impact)
    descent.set_time_options(initial_bounds=(.5, 100), duration_bounds=(.5, 100),
                             duration_ref=10, units='s')

    descent.add_state('r', units='m', rate_source='eom.r_dot', fix_initial=False, fix_final=False)
    descent.add_state('h', units='m', rate_source='eom.h_dot', targets=['atmos.h'],
                      fix_initial=False, fix_final=True)
    descent.add_state('gam', units='deg', rate_source='eom.gam_dot', targets=['eom.gam'],
                      fix_initial=False, fix_final=False)
    descent.add_state('v', units='m/s', rate_source='eom.v_dot',
                      targets=['dynamic_pressure.v', 'eom.v'],
                      fix_initial=False, fix_final=False,
                      ref=100, defect_ref=100)

    descent.add_parameter('S', targets=['aero.S'], units='m**2')
    descent.add_parameter('mass', targets=['eom.m'], units='kg')

    return descent


def new_water_rocket_trajectory(objective):
    tx_prop = dm.Radau(num_segments=50, order=3, compressed=True)
    tx_bal = dm.Radau(num_segments=10, order=3, compressed=True)
    tx_desc = dm.Radau(num_segments=10, order=3, compressed=True)
    traj = dm.Trajectory()

    # Add phases to trajectory
    propelled_ascent = traj.add_phase('propelled_ascent', new_propelled_ascent_phase(tx_prop))
    ballistic_ascent = traj.add_phase('ballistic_ascent', new_ballistic_ascent_phase(tx_bal))
    descent = traj.add_phase('descent', new_descent_phase(tx_desc))

    # Link phases
    traj.link_phases(phases=['propelled_ascent', 'ballistic_ascent'], vars=['*'])
    traj.link_phases(phases=['ballistic_ascent', 'descent'], vars=['*'])

    # Set objective function
    if objective == 'height':
        ballistic_ascent.add_objective('h', loc='final', ref=-1.0)
    elif objective == 'range':
        descent.add_objective('r', loc='final', ref=-100.0)
    else:
        raise ValueError(f"objective='{objective}' is not defined. Try using 'height' or 'range'")

    # Add design parameters to the trajectory.
    traj.add_parameter('CD',
                       targets={'propelled_ascent': ['aero.CD'],
                                'ballistic_ascent': ['aero.CD'],
                                'descent': ['aero.CD']},
                       val=0.3450, units=None, opt=False)
    traj.add_parameter('CL',
                       targets={'propelled_ascent': ['aero.CL'],
                                'ballistic_ascent': ['aero.CL'],
                                'descent': ['aero.CL']},
                       val=0.0, units=None, opt=False)
    traj.add_parameter('T',
                       targets={'ballistic_ascent': ['eom.T'],
                                'descent': ['eom.T']},
                       val=0.0, units='N', opt=False)
    traj.add_parameter('alpha',
                       targets={'propelled_ascent': ['eom.alpha'],
                                'ballistic_ascent': ['eom.alpha'],
                                'descent': ['eom.alpha']},
                       val=0.0, units='deg', opt=False)

    traj.add_parameter('m_empty', units='kg', val=0.15,
                       targets={'propelled_ascent': 'm_empty',
                                'ballistic_ascent': 'm_empty',
                                'descent': 'mass'},
                       lower=0.001, upper=1, ref=0.1,
                       opt=False)
    traj.add_parameter('V_b', units='m**3', val=2e-3,
                       targets={'propelled_ascent': 'V_b'},
                       opt=False)

    traj.add_parameter('S', units='m**2', val=np.pi*106e-3**2/4, opt=False)
    traj.add_parameter('A_out', units='m**2', val=np.pi*22e-3**2/4.,
                       targets={'propelled_ascent': ['water_engine.A_out']},
                       opt=False)
    traj.add_parameter('k', units=None, val=1.2, opt=False,
                       targets={'propelled_ascent': ['water_engine.k']})

    # Make parameter m_empty a design variable
    traj.set_parameter_options('m_empty', opt=True)

    return traj, {'propelled_ascent': propelled_ascent,
                  'ballistic_ascent': ballistic_ascent,
                  'descent': descent}


def set_sane_initial_guesses(problem, phases):
    p = problem
    # Set Initial Guesses
    p.set_val('traj.propelled_ascent.t_initial', 0.0)
    p.set_val('traj.propelled_ascent.t_duration', 0.3)

    p.set_val('traj.propelled_ascent.states:r',
              phases['propelled_ascent'].interp('r', [0, 3]))
    p.set_val('traj.propelled_ascent.states:h',
              phases['propelled_ascent'].interp('h', [0, 10]))
    # Set initial value for velocity as non-zero to avoid undefined EOM
    p.set_val('traj.propelled_ascent.states:v',
              phases['propelled_ascent'].interp('v', [0.1, 100]))
    p.set_val('traj.propelled_ascent.states:gam',
              phases['propelled_ascent'].interp('gam', [80, 80]),
              units='deg')
    p.set_val('traj.propelled_ascent.states:V_w',
              phases['propelled_ascent'].interp('V_w', [9, 0]),
              units='L')
    p.set_val('traj.propelled_ascent.states:p',
              phases['propelled_ascent'].interp('p', [6.5, 3.5]),
              units='bar')

    p.set_val('traj.ballistic_ascent.t_initial', 0.3)
    p.set_val('traj.ballistic_ascent.t_duration', 5)

    p.set_val('traj.ballistic_ascent.states:r',
              phases['ballistic_ascent'].interp('r', [0, 10]))
    p.set_val('traj.ballistic_ascent.states:h',
              phases['ballistic_ascent'].interp('h', [10, 100]))
    p.set_val('traj.ballistic_ascent.states:v',
              phases['ballistic_ascent'].interp('v', [60, 20]))
    p.set_val('traj.ballistic_ascent.states:gam',
              phases['ballistic_ascent'].interp('gam', [80, 0]),
              units='deg')

    p.set_val('traj.descent.t_initial', 10.0)
    p.set_val('traj.descent.t_duration', 10.0)

    p.set_val('traj.descent.states:r',
              phases['descent'].interp('r', [10, 20]))
    p.set_val('traj.descent.states:h',
              phases['descent'].interp('h', [10, 0]))
    p.set_val('traj.descent.states:v',
              phases['descent'].interp('v', [20, 60]))
    p.set_val('traj.descent.states:gam',
              phases['descent'].interp('gam', [0, -45]),
              units='deg')



import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

import dymos as dm
from dymos import Trajectory, GaussLobatto, Phase, Radau


class crtbp_ode(om.ExplicitComponent):
    """
    ODE for the circular restricted three-body problem (CRTBP)
    """
    def initialize(self):
        self.options.declare('num_nodes', types=int)
        self.options.declare('system', default='earth-moon', values=['earth-moon', 'sun-earth', 'jupiter-europa'],
                             desc='primary and secondary masses to be considered')

    def setup(self):
        nn = self.options['num_nodes']
        mu_val = 123.0

        self.add_input('mu', val=mu_val, desc='gravitational parameter for the specified CRTBP system')
        self.add_input('x', val=np.ones(nn), desc='x-position in rotating frame')
        self.add_input('y', val=np.ones(nn), desc='y-position in rotating frame')
        self.add_input('z', val=np.ones(nn), desc='z-position in rotating frame')
        self.add_input('x_dot', val=np.ones(nn), desc='x-velocity in rotating frame')
        self.add_input('y_dot', val=np.ones(nn), desc='y-velocity in rotating frame')
        self.add_input('z_dot', val=np.ones(nn), desc='z-velocity in rotating frame')

        self.add_output('vx', val=np.ones(nn), desc='computed velocity in the rotating frame')
        self.add_output('vy', val=np.ones(nn), desc='computed velocity in the rotating frame')
        self.add_output('vz', val=np.ones(nn), desc='computed velocity in the rotating frame')
        self.add_output('vx_dot', val=np.ones(nn), desc='computed acceleration in the rotating frame')
        self.add_output('vy_dot', val=np.ones(nn), desc='computed acceleration in the rotating frame')
        self.add_output('vz_dot', val=np.ones(nn), desc='computed acceleration in the rotating frame')


@require_pyoptsparse(optimizer='SLSQP')
def make_problem(transcription=GaussLobatto, num_segments=10, order=3, compressed=True):
    p = om.Problem(model=om.Group())
    p.driver = om.pyOptSparseDriver()
    p.driver.declare_coloring()

    traj = p.model.add_subsystem('traj', Trajectory())
    phase = traj.add_phase('phase', Phase(ode_class=crtbp_ode, transcription=transcription(num_segments=num_segments,
                                                                                           order=order,
                                                                                           compressed=compressed)))

    phase.set_time_options(fix_initial=True, fix_duration=True)
    phase.add_state('x', rate_source='vx', )
    phase.add_state('y', rate_source='vz', fix_initial=True)
    phase.add_state('z', rate_source='vz')
    phase.add_state('x_dot', rate_source='vx_dot', fix_initial=True, units=None)
    phase.add_state('y_dot', rate_source='vy_dot', units=None)
    phase.add_state('z_dot', rate_source='vz_dot', fix_initial=True, units=None)

    p.model.add_subsystem('x_periodic_bc', om.ExecComp('bc_defect=final-initial'))
    p.model.connect('traj.phase.timeseries.states:x', 'x_periodic_bc.initial', src_indices=0, flat_src_indices=True)
    p.model.connect('traj.phase.timeseries.states:x', 'x_periodic_bc.final', src_indices=-1, flat_src_indices=True)

    p.model.add_constraint('x_periodic_bc.bc_defect', equals=0)

    p.model.add_subsystem('z_periodic_bc', om.ExecComp('bc_defect=final-initial'))
    p.model.connect('traj.phase.timeseries.states:z', 'z_periodic_bc.initial', src_indices=0, flat_src_indices=True)
    p.model.connect('traj.phase.timeseries.states:z', 'z_periodic_bc.final', src_indices=-1, flat_src_indices=True)

    p.model.add_constraint('z_periodic_bc.bc_defect', equals=0)

    p.model.add_subsystem('vy_periodic_bc', om.ExecComp('bc_defect=final-initial'))
    p.model.connect('traj.phase.timeseries.states:y_dot', 'vy_periodic_bc.initial', src_indices=0, flat_src_indices=True)
    p.model.connect('traj.phase.timeseries.states:y_dot', 'vy_periodic_bc.final', src_indices=-1, flat_src_indices=True)

    p.model.add_constraint('vy_periodic_bc.bc_defect', equals=0)

    phase.add_objective('time', loc='final')

    p.setup(check=True)
    return p


@use_tempdirs
class TestPseudospectralBase(unittest.TestCase):

    def test_add_state_units_none(self):
        p = make_problem(transcription=GaussLobatto, num_segments=10, order=3, compressed=True)
        p.run_model()

        io_meta = p.model.traj.phases.phase.timeseries.get_io_metadata(iotypes=('output'), get_remote=True)
        self.assertEqual(io_meta['states:x']['units'], None)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import numpy as np

import openmdao.api as om


class KappaComp(om.ExplicitComponent):
    r""" Computes the term kappa in the drag equation:

    .. math::

        C_D = C_{D0} + \kappa C_{L\alpha} \alpha^2

    """
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # Inputs
        self.add_input('mach', shape=(nn,), desc='Mach number', units=None)

        # Outputs
        self.add_output(name='kappa', val=np.zeros(nn), desc='induced drag coefficient', units=None)

        # Jacobian
        ar = np.arange(nn)
        self.declare_partials(of='kappa', wrt='mach', rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        M = inputs['mach']

        idx_low = np.where(M < 1.15)[0]
        idx_high = np.where(M >= 1.15)[0]

        outputs['kappa'][idx_low] = 0.54 + 0.15 * (1.0 + np.tanh((M[idx_low] - 0.9) / 0.06))
        outputs['kappa'][idx_high] = 0.54 + 0.15 * (1.0 + np.tanh(0.25 / 0.06)) \
            + 0.14 * (M[idx_high] - 1.15)

    def compute_partials(self, inputs, partials):
        M = inputs['mach']

        idx_low = np.where(M < 1.15)[0]
        idx_high = np.where(M >= 1.15)[0]

        k = 50.0 / 3.0
        tanh = np.tanh(k * (M[idx_low] - 0.9))
        sech2 = 1.0 - tanh**2

        partials['kappa', 'mach'][idx_low] = 2.5 * sech2
        partials['kappa', 'mach'][idx_high] = 0.14



from setuptools import find_packages, setup

# Setup optional dependencies
optional_dependencies = {
    'docs': [
        'matplotlib',
        'jupyter',
        'jupyter-book',
        'nbconvert',
        'notebook',
        'ipython',
        'numpydoc>=1.1',
        'redbaron',
        'tabulate',
        'jaxlib',
        'jax'
    ],
    'notebooks': [
        'notebook',
        'tabulate',
        'ipython'
    ],
    'test': [
        'pep8',
        'testflo>=1.3.6',
        'matplotlib',
        'numpydoc>=1.1',
        'playwright>=1.20',
        'aiounittest'
    ]
}

# Add an optional dependency that concatenates all others
optional_dependencies['all'] = sorted([
    dependency
    for dependencies in optional_dependencies.values()
    for dependency in dependencies
])


setup(name='dymos',
    version='1.6.0',
    description='Open-Source Optimization of Dynamic Multidisciplinary Systems',
    long_description='''
Dymos is a framework for the simulation and optimization of dynamical systems within the OpenMDAO Multidisciplinary Analysis and Optimization environment.
Dymos leverages implicit and explicit simulation techniques to simulate generic dynamic systems of arbitary complexity.

The software has two primary objectives:
-   Provide a generic ODE integration interface that allows for the analysis of dynamical systems.
-   Allow the user to solve optimal control problems involving dynamical multidisciplinary systems.''',
    long_description_content_type='text/markdown',
    url='https://github.com/OpenMDAO/dymos',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft :: Windows',
        'Topic :: Scientific/Engineering',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    license='Apache License',
    packages=find_packages(),
    python_requires=">=3.8",
    install_requires=[
        'openmdao>=3.17.0',
        'numpy',
        'scipy'
    ],
    extras_require=optional_dependencies,
    zip_safe=False,
    package_data={
        'dymos.examples.aircraft_steady_flight.aero': ['data/CRM_aero_inputs.dat', 'data/CRM_aero_outputs.dat'],
        'dymos.visualization.linkage': ['report_template.html', 'js/*', 'style/*']
    },
    entry_points = {
        'openmdao_report': [
            'dymos.linkage=dymos.visualization.linkage.report:_linkage_report_register'
        ],
    },
)



import unittest
import numpy as np
from numpy.polynomial import Polynomial as P
import matplotlib.pyplot as plt
import matplotlib.cm as cm

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from dymos.utils.testing_utils import assert_timeseries_near_equal
import dymos as dm
from dymos.examples.min_time_climb.min_time_climb_ode import MinTimeClimbODE

from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse


def min_time_climb(optimizer='SLSQP', num_seg=3, transcription='gauss-lobatto',
                   transcription_order=3, force_alloc_complex=False, add_rate=False):

    p = om.Problem(model=om.Group())

    p.driver = om.pyOptSparseDriver()
    p.driver.options['optimizer'] = optimizer
    p.driver.declare_coloring()

    if optimizer == 'SNOPT':
        p.driver.opt_settings['Major iterations limit'] = 1000
        p.driver.opt_settings['iSumm'] = 6
        p.driver.opt_settings['Major feasibility tolerance'] = 1.0E-6
        p.driver.opt_settings['Major optimality tolerance'] = 1.0E-6
        p.driver.opt_settings['Function precision'] = 1.0E-12
        p.driver.opt_settings['Linesearch tolerance'] = 0.1
        p.driver.opt_settings['Major step limit'] = 0.5
    elif optimizer == 'IPOPT':
        p.driver.opt_settings['tol'] = 1.0E-5
        p.driver.opt_settings['print_level'] = 0
        p.driver.opt_settings['mu_strategy'] = 'monotone'
        p.driver.opt_settings['bound_mult_init_method'] = 'mu-based'
        p.driver.opt_settings['mu_init'] = 0.01

    t = {'gauss-lobatto': dm.GaussLobatto(num_segments=num_seg, order=transcription_order),
         'radau-ps': dm.Radau(num_segments=num_seg, order=transcription_order)}

    traj = dm.Trajectory()

    phase = dm.Phase(ode_class=MinTimeClimbODE, transcription=t[transcription])
    traj.add_phase('phase0', phase)

    p.model.add_subsystem('traj', traj)

    phase.set_time_options(fix_initial=True, duration_bounds=(50, 400),
                           duration_ref=100.0)

    phase.add_state('r', fix_initial=True, lower=0, upper=1.0E6,
                    ref=1.0E3, defect_ref=1.0E3, units='m',
                    rate_source='flight_dynamics.r_dot')

    phase.add_state('h', fix_initial=True, lower=0, upper=20000.0,
                    ref=20_000, defect_ref=20_000, units='m',
                    rate_source='flight_dynamics.h_dot', targets=['h'])

    phase.add_state('v', fix_initial=True, lower=10.0,
                    ref=1.0E2, defect_ref=1.0E2, units='m/s',
                    rate_source='flight_dynamics.v_dot', targets=['v'])

    phase.add_state('gam', fix_initial=True, lower=-1.5, upper=1.5,
                    ref=1.0, defect_ref=1.0, units='rad',
                    rate_source='flight_dynamics.gam_dot', targets=['gam'])

    phase.add_state('m', fix_initial=True, lower=10.0, upper=1.0E5,
                    ref=10_000, defect_ref=10_000, units='kg',
                    rate_source='prop.m_dot', targets=['m'])

    phase.add_control('alpha', units='deg', lower=-8.0, upper=8.0, scaler=1.0,
                      rate_continuity=True, rate_continuity_scaler=100.0,
                      rate2_continuity=False, targets=['alpha'])

    phase.add_parameter('S', val=49.2386, units='m**2', opt=False, targets=['S'])
    phase.add_parameter('Isp', val=1600.0, units='s', opt=False, targets=['Isp'])
    phase.add_parameter('throttle', val=1.0, opt=False, targets=['throttle'])

    phase.add_boundary_constraint('h', loc='final', equals=20000, scaler=1.0E-3)
    phase.add_boundary_constraint('aero.mach', loc='final', equals=1.0)
    phase.add_boundary_constraint('gam', loc='final', equals=0.0)

    phase.add_path_constraint(name='h', lower=100.0, upper=20000, ref=20000)
    phase.add_path_constraint(name='aero.mach', lower=0.1, upper=1.8)

    # Unnecessary but included to test capability
    phase.add_path_constraint(name='alpha', lower=-8, upper=8)
    phase.add_path_constraint(name='time', lower=0, upper=400)
    phase.add_path_constraint(name='time_phase', lower=0, upper=400)

    # Minimize time at the end of the phase
    phase.add_objective('time', loc='final', ref=1.0)

    # test mixing wildcard ODE variable expansion and unit overrides
    phase.add_timeseries_output(['aero.*', 'prop.thrust', 'prop.m_dot'],
                                units={'aero.f_lift': 'lbf', 'prop.thrust': 'lbf'})

    # test adding rate as timeseries output
    if add_rate:
        phase.add_timeseries_rate_output('aero.mach')

    p.model.linear_solver = om.DirectSolver()

    p.setup(check=True, force_alloc_complex=force_alloc_complex)

    p['traj.phase0.t_initial'] = 0.0
    p['traj.phase0.t_duration'] = 350.0

    p['traj.phase0.states:r'] = phase.interp('r', [0.0, 111319.54])
    p['traj.phase0.states:h'] = phase.interp('h', [100.0, 20000.0])
    p['traj.phase0.states:v'] = phase.interp('v', [135.964, 283.159])
    p['traj.phase0.states:gam'] = phase.interp('gam', [0.0, 0.0])
    p['traj.phase0.states:m'] = phase.interp('m', [19030.468, 16841.431])
    p['traj.phase0.controls:alpha'] = phase.interp('alpha', [0.0, 0.0])

    dm.run_problem(p, simulate=True)

    return p


@use_tempdirs
class TestMinTimeClimb(unittest.TestCase):

    def _test_results(self, p):
        """ Verify the results of the optimization. """
        # Verify that ODE output mach is added to the timeseries
        assert_near_equal(p.get_val('traj.phase0.timeseries.mach')[-1], 1.0, tolerance=1.0E-2)

        # Check that time matches to within 1% of an externally verified solution.
        assert_near_equal(p.get_val('traj.phase0.timeseries.time')[-1], 321.0, tolerance=0.02)

    def _test_wilcard_outputs(self, p):
        """ Test that all wilcard outputs are provided. """
        output_dict = dict(p.model.list_outputs(units=True, out_stream=None))
        ts = {k: v for k, v in output_dict.items() if 'timeseries.' in k}
        for c in ['mach', 'CD0', 'kappa', 'CLa', 'CL', 'CD', 'q', 'f_lift', 'f_drag', 'thrust', 'm_dot']:
            assert(any([True for t in ts if 'timeseries.' + c in t]))

    def _test_timeseries_units(self, p):
        """ Test that the units from the timeseries are correct. """
        output_dict = dict(p.model.list_outputs(units=True, out_stream=None))
        assert(output_dict['traj.phases.phase0.timeseries.thrust']['units'] == 'lbf')  # no wildcard, from units dict
        assert(output_dict['traj.phases.phase0.timeseries.m_dot']['units'] == 'kg/s')  # no wildcard, from ODE
        assert(output_dict['traj.phases.phase0.timeseries.f_drag']['units'] == 'N')    # wildcard, from ODE
        assert(output_dict['traj.phases.phase0.timeseries.f_lift']['units'] == 'lbf')  # wildcard, from units dict

    def _test_mach_rate(self, p, plot=False):
        """ Test that the mach rate is provided by the timeseries and is accurate. """
        # Verify correct timeseries output of mach_rate
        output_dict = dict(p.model.list_outputs(units=True, out_stream=None))
        ts = {k: v for k, v in output_dict.items() if 'timeseries.' in k}
        self.assertTrue('traj.phases.phase0.timeseries.mach_rate' in ts)

        case = om.CaseReader('dymos_solution.db').get_case('final')

        time = case['traj.phases.phase0.timeseries.time'][:, 0]
        mach = case['traj.phases.phase0.timeseries.mach'][:, 0]
        mach_rate = case['traj.phases.phase0.timeseries.mach_rate'][:, 0]

        sim_case = om.CaseReader('dymos_simulation.db').get_case('final')

        sim_time = sim_case['traj.phases.phase0.timeseries.time'][:, 0]
        sim_mach = sim_case['traj.phases.phase0.timeseries.mach'][:, 0]
        sim_mach_rate = sim_case['traj.phases.phase0.timeseries.mach_rate'][:, 0]

        # Fit a numpy polynomial segment by segment to mach vs time, and compare the derivatives to mach_rate
        gd = p.model.traj.phases.phase0.options['transcription'].grid_data
        seg_idxs = gd.subset_segment_indices['all']
        num_seg = seg_idxs.shape[0]

        sim_seg_idxs = np.zeros_like(seg_idxs)
        sim_seg_idxs[:, 0] = np.arange(0, 10*num_seg, 10, dtype=int)
        sim_seg_idxs[:, 1] = sim_seg_idxs[:, 0] + 10

        if plot:
            fig, axes = plt.subplots(2, 1, sharex=True)
            color = iter(cm.viridis(np.linspace(0, 1, num_seg)))
            axes[0].set_ylabel('Mach')
            axes[1].set_ylabel('Mach rate (1/s)')
            axes[1].set_xlabel('time (s)')

        for i in range(num_seg):
            time_seg = time[seg_idxs[i, 0]: seg_idxs[i, 1]]
            mach_seg = mach[seg_idxs[i, 0]: seg_idxs[i, 1]]
            order = len(time_seg) - 1
            p = P.fit(time_seg, mach_seg, order)
            deriv = p.deriv(1)

            time_nodes = time[seg_idxs[i, 0]: seg_idxs[i, 1]]
            mach_nodes = mach[seg_idxs[i, 0]: seg_idxs[i, 1]]
            mach_rate_nodes = mach_rate[seg_idxs[i, 0]: seg_idxs[i, 1]]

            if plot:
                c = next(color)
                t_plot = np.linspace(time_seg[0], time_seg[-1], 20)
                m_plot = p(t_plot)
                m_rate_plot = deriv(t_plot)
                axes[0].plot(t_plot, m_plot, c=c)
                axes[1].plot(t_plot, m_rate_plot, '--', c=c)

                axes[0].plot(time[seg_idxs[i, 0]: seg_idxs[i, 1]], mach[seg_idxs[i, 0]: seg_idxs[i, 1]], 'o', c=c)
                axes[1].plot(time[seg_idxs[i, 0]: seg_idxs[i, 1]], mach_rate[seg_idxs[i, 0]: seg_idxs[i, 1]], 'o', c=c)

                axes[0].plot(sim_time[sim_seg_idxs[i, 0]: sim_seg_idxs[i, 1]],
                             sim_mach[sim_seg_idxs[i, 0]: sim_seg_idxs[i, 1]], ':', c=c, lw=2)
                axes[1].plot(sim_time[sim_seg_idxs[i, 0]: sim_seg_idxs[i, 1]],
                             sim_mach_rate[sim_seg_idxs[i, 0]: sim_seg_idxs[i, 1]], ':', c=c, lw=2)

            assert_near_equal(mach_nodes, p(time_nodes), tolerance=1.0E-9)
            assert_near_equal(mach_rate_nodes, deriv(time_nodes), tolerance=1.0E-9)

        # Comparing the mach rate over the entire trajectory since it is expected to be off at some points due to
        # the equidistant time-spacing of nodes in SolveIVP's timeseries outputs.
        assert_timeseries_near_equal(t_ref=time, x_ref=mach_rate, t_check=sim_time, x_check=sim_mach_rate,
                                     abs_tolerance=0.02, rel_tolerance=0.02)

        if plot:
            plt.show()

    @require_pyoptsparse(optimizer='SLSQP')
    def test_results_gauss_lobatto(self):
        NUM_SEG = 12
        ORDER = 3
        p = min_time_climb(optimizer='IPOPT', num_seg=NUM_SEG, transcription_order=ORDER,
                           transcription='gauss-lobatto', add_rate=True)

        self._test_results(p)

        self._test_wilcard_outputs(p)

        self._test_timeseries_units(p)

        self._test_mach_rate(p)

    @require_pyoptsparse(optimizer='SLSQP')
    def test_results_radau(self):
        NUM_SEG = 15
        ORDER = 3
        p = min_time_climb(optimizer='SLSQP', num_seg=NUM_SEG, transcription_order=ORDER,
                           transcription='radau-ps', add_rate=True)

        self._test_results(p)

        self._test_wilcard_outputs(p)

        self._test_timeseries_units(p)

        self._test_mach_rate(p, plot=False)

if __name__ == '__main__':  # pragma: no cover
    unittest.main()



from .continuity_comp import RadauPSContinuityComp, GaussLobattoContinuityComp
from .control_group import ControlGroup
from .parameter_comp import ParameterComp
from .polynomial_control_group import PolynomialControlGroup
from .time_comp import TimeComp



import numpy as np


class LagrangeBarycentricInterpolant(object):
    """
    Class definition for LagrangeBarycentricInterpolant.

    Interpolates values and first derivatives of set of data using barycentric interpolation of a
    Lagrange Polynomial.

    Parameters
    ----------
    nodes : sequence
        The nodes of the polynomial from -1 to 1 on which the values to
        be interpolated are given.
    shape : tuple
        Shape of the values to be interpolated.

    Attributes
    ----------
    num_nodes : int
        The number of nodes in the interpolated polynomial.
    tau_i : np.array
        The location of the nodes of the polynomial in tau space [-1 1].
    x_0 : float
        The x-value corresponding to the left-side of the interval (tau = -1).
    x_f : float
        The x-value corresponding to the right-side of the interval (tau = +1).
    f_j : np.array
        The values to be interpolated
    w_b : np.array
        The barycentric weights for the points in the interpolated polynomial.
    wbfj : np.array
        An array of the precomputed product of the interpolated values and
        the corresponding barycentric weights.
    dx_dtau : float
        Half the span from x0 to xf.  The ratio of x-space to
        internal tau-space.

    Notes
    -----
    The Barycentric formula is given in Eq. 3.3 of [1]_ as

    .. math::

        p(x) = l(x) \\Sum \\frac{w_j f_j}{x-x_j}

    where l(x) is

    .. math::

        l(x) = (x-x_0)(x-x_1)(x-x_2)...

    The singularity in the denominator of p(x) at x = x_n is cancelled
    by the the term (x-x_n) in l(x).

    References
    ----------
    .. [1] Berrut, Jean-Paul, and Lloyd N. Trefethen.
       "Barycentric lagrange interpolation." Siam Review 46.3 (2004): 501-517.
    """

    def __init__(self, nodes, shape):

        self.num_nodes = len(nodes)
        """ The number of nodes in the interpolated polynomial. """

        self.tau_i = nodes
        """ The independent variable values at interpolation points. """

        _shape = (self.num_nodes,) + shape

        self.f_j = np.zeros(_shape)
        """ An array of values to be interpolated. """

        self.w_b = np.ones(self.num_nodes)
        """ Barycentric weights for nodes in the interpolated polynomial."""

        # Barycentric Weights
        for j in range(self.num_nodes):
            for k in range(self.num_nodes):
                if k != j:
                    self.w_b[j] /= (self.tau_i[j] - self.tau_i[k])

        self.wbfj = np.zeros(_shape)
        """ An array of the precomputed product of the interpolated
            values and the corresponding barycentric weights.
        """

        n = self.wbfj.shape[0]
        m = np.prod(self.wbfj.shape[1:])
        self.wbfj_flat = np.reshape(self.wbfj, newshape=(n, m))
        """ A flattened view of wbfj"""

        self.x0 = -1.0
        """ The value of the independent axis corresponding to $\tau = -1$ """

        self.xf = 1.0
        """ The value of the independent axis corresponding to $\tau = 1$ """

        self.dx_dtau = 1.0
        """ Half the span from x0 to xf.  The ratio of x-space to
        internal tau-space. """

        self._is_setup = False

    def x_to_tau(self, x):
        """
        Converts the independent variable x to its corresponding value of $\tau$.

        Given bounds on the independent variable x0 and xf which
        correspond to $\tau$ of -1 and 1, respectively, the returned value
        will be the equivalent $tau$.

        For instance, if x0 = 0 and xf = 100, x = 50 will have an equivalent
        $\tau$ of 0 (halfway on [-1 1]).

        Parameters
        ----------
        x : float or ndarray
            The independent variable to be converted to $\tau$.

        Returns
        -------
        float or ndarray
            The equivalent value of $\tau$ given x.
        """
        return -1.0 + (x - self.x0) / self.dx_dtau

    def setup(self, x0, xf, f_j):
        """
        Prepare the interpolant for use by setting the values to be interpolated.

        Parameters
        ----------
        x0 : float
            The lower bound of the independent variable.
            corresponding to $\tau = -1$ .
        xf : float
            The upper bound of the independent variable.
            corresponding to $\tau = -1$ .
        f_j : np.array
            The values to be interpolated at the LGL nodes on [-1 1].

        Raises
        ------
        ValueError
            If the length of f_j is not the number of nodes in the interpolated
            polynomial.
        """
        if len(f_j) != self.num_nodes:
            raise ValueError(f"f_j must have {self.num_nodes} values")
        self.f_j[...] = f_j
        self.x0 = x0
        self.xf = xf
        self.dx_dtau = 0.5 * (xf - x0)

        fjT = self.f_j.T
        self.wbfj[...] = (self.w_b * fjT).T
        self._is_setup = True

    def eval(self, x):
        """
        Interpolate the LGL polynomial at x.

        Parameters
        ----------
        x : float
            The independent variable value at which interpolation
            is requested.

        Returns
        -------
        float
            The interpolated value of the LGL polynomial at x.
        """
        if not self._is_setup:
            raise RuntimeError('LagrangeBarycentricInterpolant has not been setup')
        tau = self.x_to_tau(x)

        g = tau - self.tau_i
        l = np.ones_like(g)

        for i in range(self.num_nodes):
            for j in range(self.num_nodes):
                if j != i:
                    l[i] *= g[j]

        result = np.reshape(np.dot(l, self.wbfj_flat), newshape=self.wbfj.shape[1:])

        return result

    def eval_deriv(self, x, der=1):
        """
        Interpolate the derivative of the polynomial at x.

        Parameters
        ----------
        x : float
            The independent variable value at which the derivative
            is requested.
        der : int
            Derivative order requested. Default is 1 for first derivatives.

        Returns
        -------
        float
            The first derivative of the polynomial at x.
        """
        if not self._is_setup:
            raise RuntimeError('LagrangeBarycentricInterpolant has not been setup')
        if der >= self.num_nodes:
            return 0.0

        n = self.num_nodes
        tau = self.x_to_tau(x)
        g = tau - self.tau_i
        lprime = np.zeros(n)

        if der == 1:
            for i in range(n):
                for j in range(n):
                    if j == i:
                        continue
                    prod = 1.0
                    for k in range(n):
                        if k != i and k != j:
                            prod *= g[k]
                    lprime[i] += prod
            # df_dtau = np.dot(lprime, self.wbfj)
            df_dtau = np.reshape(np.dot(lprime, self.wbfj_flat), newshape=self.wbfj.shape[1:])
            return df_dtau / self.dx_dtau
        elif der == 2:
            for i in range(n):
                for j in range(n):
                    if j == i:
                        continue
                    for k in range(n):
                        if k == i or k == j:
                            continue
                        prod = 1.0
                        for ii in range(n):
                            if ii != i and ii != j and ii != k:
                                prod *= g[ii]
                        lprime[i] += prod
            df_dtau = np.reshape(np.dot(lprime, self.wbfj_flat), newshape=self.wbfj.shape[1:])
            return df_dtau / self.dx_dtau**2
        else:
            raise ValueError('Barycentric interpolant currently only supports up to '
                             'second derivatives')



import numpy as np
import openmdao.api as om


class HyperSensitiveODE(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # inputs
        self.add_input('x', val=np.zeros(nn), desc='state')
        self.add_input('xL', val=np.zeros(nn), desc='cost_state')

        self.add_input('u', val=np.zeros(nn), desc='control')

        self.add_output('x_dot', val=np.zeros(nn), desc='state rate', units='1/s')
        self.add_output('L', val=np.zeros(nn), desc='Lagrangian', units='1/s')

        # Setup partials
        self.declare_partials(of='x_dot', wrt='x', rows=np.arange(nn), cols=np.arange(nn), val=-1)
        self.declare_partials(of='x_dot', wrt='u', rows=np.arange(nn), cols=np.arange(nn), val=1)

        self.declare_partials(of='L', wrt='x', rows=np.arange(nn), cols=np.arange(nn))
        self.declare_partials(of='L', wrt='u', rows=np.arange(nn), cols=np.arange(nn))

    def compute(self, inputs, outputs):
        x = inputs['x']
        u = inputs['u']

        outputs['x_dot'] = -x + u
        outputs['L'] = (x ** 2 + u ** 2) / 2

    def compute_partials(self, inputs, jacobian):
        x = inputs['x']
        u = inputs['u']

        jacobian['L', 'x'] = x
        jacobian['L', 'u'] = u



import numpy as np
import openmdao.api as om
import scipy.sparse as sp

from openmdao.utils.units import unit_conversion
from dymos.transcriptions.common.timeseries_output_comp import TimeseriesOutputCompBase

from ....utils.lagrange import lagrange_matrices


class SolveIVPTimeseriesOutputComp(TimeseriesOutputCompBase):
    """
    Class definition for SolveIVPTimeseriesOutputComp.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def initialize(self):
        """
        Declare component options.
        """
        super(SolveIVPTimeseriesOutputComp, self).initialize()

        self.options.declare('output_nodes_per_seg', default=None, types=(int,), allow_none=True,
                             desc='If None, results are provided at the all nodes within each'
                                  'segment.  If an int (n) then results are provided at n '
                                  'equally distributed points in time within each segment.')

        self.options.declare('time_units', default=None, allow_none=True, types=str,
                             desc='Units of time')

    def setup(self):
        """
        Define the independent variables as output variables.
        """
        igd = self.options['input_grid_data']

        if self.options['output_nodes_per_seg'] is None:
            self.output_num_nodes = self.input_num_nodes = igd.num_nodes
            ogd = igd
        else:
            self.output_num_nodes = self.input_num_nodes = igd.num_segments * self.options['output_nodes_per_seg']
            output_nodes_stau = np.linspace(-1, 1, self.options['output_nodes_per_seg'])
            output_nodes_ptau = igd.node_ptau[igd.subset_node_indices['all']]
            ogd = None

        # Build the interpolation matrix which maps from the input grid to the output grid.
        # Rather than a single phase-wide interpolating polynomial, map each segment.
        # To do this, find the nodes in the output grid which fall in each segment of the input
        # grid.  Then build a Lagrange interpolating polynomial for that segment
        # L_blocks = []
        D_blocks = []

        for iseg in range(igd.num_segments):
            # Create the interpolation matrix and add it to the blocks
            if self.options['output_nodes_per_seg'] is None:
                i1, i2 = igd.segment_indices[iseg]
                iptau_segi = igd.node_ptau[i1:i2]
                istau_segi = igd.node_stau[i1:i2]

                # The indices of the output grid that fall within this segment of the input grid
                if ogd is igd:
                    optau_segi = iptau_segi
                else:
                    ptau_hi = igd.segment_ends[iseg + 1]
                    if iseg < igd.num_segments - 1:
                        optau_segi = output_nodes_ptau[output_nodes_ptau <= ptau_hi]
                    else:
                        optau_segi = output_nodes_ptau

                    # Remove the captured nodes so we don't accidentally include them again
                    output_nodes_ptau = output_nodes_ptau[len(optau_segi):]

                # # Now get the output nodes which fall in iseg in iseg's segment tau space.
                ostau_segi = 2.0 * (optau_segi - iptau_segi[0]) / (iptau_segi[-1] - iptau_segi[0]) - 1

                _, D = lagrange_matrices(istau_segi, ostau_segi)
            else:
                _, D = lagrange_matrices(output_nodes_stau, output_nodes_stau)
            D_blocks.append(D)

        self.differentiation_matrix = sp.block_diag(D_blocks, format='csr')

        for (name, kwargs) in self._vars:
            units = kwargs['units']
            desc = kwargs['desc']
            shape = kwargs['shape']
            src = kwargs['src']
            rate = kwargs['rate']
            self._add_output_configure(name, units, shape, desc, src, rate)

        self.add_input('dt_dstau', shape=(self.output_num_nodes,), units=self.options['time_units'])

    def _add_output_configure(self, name, units, shape, desc, src=None, rate=False):
        """
        Add a single timeseries output.

        Can be called by parent groups in configure.

        Parameters
        ----------
        name : str
            name of the variable in this component's namespace.
        shape : int or tuple or list or None
            Shape of this variable, only required if val is not an array.
            Default is None.
        units : str or None
            Units in which the output variables will be provided to the component during execution.
            Default is None, which means it has no units.
        desc : str
            description of the timeseries output variable.
        src : str or None
            The source of the timeseries output.
        rate : bool
            If True, timeseries output is a rate.
        """
        if rate:
            om.issue_warning(f'Timeseries rate outputs not currently supported in simulate: {name} being skipped.')

        input_num_nodes = self.input_num_nodes
        output_num_nodes = self.output_num_nodes
        added_source = False

        if name in self._vars:
            return False

        if src in self._sources:
            # If we're already pulling the source into this timeseries, use that as the
            # input for this output.
            input_name = self._sources[src]
            input_units = self._units[input_name]
        else:
            input_name = f'input_values:{name}'
            self.add_input(input_name,
                           shape=(input_num_nodes,) + shape,
                           units=units, desc=desc)
            self._sources[src] = input_name
            input_units = self._units[input_name] = units
            added_source = True

        output_name = name
        self.add_output(output_name, shape=(output_num_nodes,) + shape, units=units, desc=desc)

        self._vars[name] = (input_name, output_name, shape, rate)

        # There's a chance that the input for this output was pulled from another variable with
        # different units, so account for that with a conversion.
        if input_units is None or units is None:
            pass
        else:
            scale, offset = unit_conversion(input_units, units)
            self._conversion_factors[output_name] = scale, offset

        return added_source

    def compute(self, inputs, outputs):
        """
        Compute component outputs.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        """
        dt_dstau = inputs['dt_dstau'][:, np.newaxis]
        for (input_name, output_name, _, is_rate) in self._vars.values():
            interp_vals = inputs[input_name]

            if is_rate:
                interp_vals = self.differentiation_matrix.dot(interp_vals) / dt_dstau

            if output_name in self._conversion_factors:
                scale, offset = self._conversion_factors[output_name]
                outputs[output_name] = scale * (interp_vals + offset)
            else:
                outputs[output_name] = interp_vals
